# -*- coding: utf-8 -*-

"""

BEWEGUNG
a versatile video renderer
https://github.com/pleiszenburg/bewegung

    src/bewegung/core/effects.py: Video frame effects

    Copyright (C) 2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/bewegung/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import math

from PIL import Image as PIL_Image

from .abc import EffectABC, LayerABC, SequenceABC, TimeABC, VideoABC
from .typeguard import typechecked

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS: BASE
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@typechecked
class EffectBase(EffectABC):
    """
    Base for all effect classes - not an effect on its own. Derive new effect classes from this class.
    Decorator for layers.

    Mutable.

    If the orginal cunstructor method is overridden, it must be called from the child class.
    """

    def __init__(self):

        apply = getattr(self.apply, '__wrapped__', self.apply) # typeguard
        self._args = apply.__code__.co_varnames[
            1:apply.__code__.co_argcount # excluding self and internal namespace
            ] # parameters requested by user
        assert self._args[0] == 'cvs' # canvas

    def __repr__(self) -> str:

        return f'<{type(self).__name__}>'

    def __call__(self, layer: LayerABC) -> LayerABC:
        """
        Decorator function, decorating :class:`bewegung.core.layer.Layer` objects (wrapping user-defined layer methods).
        The effect is registered via :meth:`bewegung.core.layer.Layer.register_effect` and returns the otherwise unchanged layer object.

        Do not override!

        Args:
            layer : Layer to which the effect is applied.
        """

        layer.register_effect(self)
        return layer

    def apply_(self,
        cvs: PIL_Image.Image,
        video: VideoABC,
        sequence: SequenceABC,
        time: TimeABC,
        ) -> PIL_Image.Image:
        """
        Internal interface for layer objects. Applies the effect to a Pillow Image object and returns the modified image.
        The function automatically determines what arguments the ``apply`` method of an actual effect requests other than ``cvs``.
        Possible options are:

        - video: Parent video object
        - sequence: Parent sequence object
        - time: Time within parent video
        - reltime: Relative time within parent sequence

        Do not override!

        Args:
            cvs : Input Pillow Image object
            video : Parent video object
            sequence : Parent sequence object
            time : Time within parent video
        """

        kwargs = {}
        for arg in self._args[1:]:
            if arg == 'video':
                kwargs['video'] = video
            elif arg == 'sequence':
                kwargs['sequence'] = sequence
            elif arg == 'time':
                kwargs['time'] = time
            elif arg == 'reltime':
                kwargs['reltime'] = time - sequence.start
            else:
                raise ValueError('unknown argument')

        return self.apply(cvs = cvs, **kwargs)

    def apply(self, cvs: PIL_Image.Image):
        """
        Applies effect to image.

        Must be reimplemented!

        Args:
            cvs : Input Pillow Image object
        """

        raise NotImplementedError()

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASSES: EFFECTS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@typechecked
class FadeInEffect(EffectBase):
    """
    Fade-in effect. Decorator for layers.

    Args:
        blend_time : Duration of effect relative to the beginning of the parent sequence
    """

    def __init__(self, blend_time: TimeABC):

        if blend_time.index < 0:
            raise ValueError('time must be positive or zero')

        super().__init__()
        self._blend_time = blend_time

    def apply(self, cvs: PIL_Image.Image, sequence: SequenceABC, time: TimeABC,) -> PIL_Image.Image:

        reltime = time - sequence.start
        if reltime > self._blend_time:
            return cvs

        r, g, b, a = cvs.split()
        factor = _sin_fade(reltime.index / self._blend_time.index)
        a = a.point(lambda i: i * factor)

        return PIL_Image.merge('RGBA', (r, g, b, a))

@typechecked
class FadeOutEffect(EffectBase):
    """
    Fade-out effect. Decorator for layers.

    Args:
        blend_time : Duration of effect relative to the beginning of the parent sequence
    """

    def __init__(self, blend_time: TimeABC):

        if blend_time.index < 0:
            raise ValueError('time must be positive or zero')

        super().__init__()
        self._blend_time = blend_time

    def apply(self, cvs: PIL_Image.Image, sequence: SequenceABC, time: TimeABC,) -> PIL_Image.Image:

        if time < sequence.stop - self._blend_time:
            return cvs
        nreltime = sequence.stop - time

        r, g, b, a = cvs.split()
        factor = _sin_fade(nreltime.index / self._blend_time.index)
        a = a.point(lambda i: i * factor)

        return PIL_Image.merge('RGBA', (r, g, b, a))

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# ROUTINES: HELPER
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@typechecked
def _sin_fade(factor: float) -> float:

    assert 0.0 <= factor <= 1.0

    return -math.cos(factor * math.pi) / 2 + 0.5



# -*- coding: utf-8 -*-

"""

BEWEGUNG
a versatile video renderer
https://github.com/pleiszenburg/bewegung

    src/bewegung/core/time.py: Time handling

    Copyright (C) 2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/bewegung/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from typing import Generator, Union

from .abc import TimeABC
from .const import FPS_DEFAULT
from .typeguard import typechecked

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@typechecked
class Time(TimeABC):
    """
    This class represents time both as number of frames (the index) and time in seconds.
    For conversion, it has an internal frames per second state.
    Operators for basic arithmetic such as add and substract are implemented.
    Comparison operators are implemented.
    Operations can only be performed on ``Time`` objects with equal frames per second.
    If frames per second are unequal, an exception will be raised.

    Immutable.

    Args:
        fps : Frames per second
        index : Number of frames
    """

    def __init__(self, fps: int = FPS_DEFAULT, index: int = 0):

        if fps <= 0:
            raise ValueError('there must be at least one frame per second')

        self._fps, self._index = fps, index

    def __repr__(self) -> str:

        return f'<Time index={self._index:d} seconds={self.seconds:.03f}s fps={self._fps:d}>'

    def __int__(self) -> int:
        """
        If converted to an integer, the frame number (index) will be exposed.
        """

        return self._index

    def __float__(self) -> float:
        """
        If converted to a floating point number, the time in seconds will be exposed.
        """

        return self.seconds

    def __eq__(self, other: TimeABC) -> bool:
        self._assert_fps(other)
        return self.index == other.index

    def __lt__(self, other: TimeABC) -> bool:
        self._assert_fps(other)
        return self.index < other.index

    def __le__(self, other: TimeABC) -> bool:
        self._assert_fps(other)
        return self.index <= other.index

    def __gt__(self, other: TimeABC) -> bool:
        self._assert_fps(other)
        return self.index > other.index

    def __ge__(self, other: TimeABC) -> bool:
        self._assert_fps(other)
        return self.index >= other.index

    def __add__(self, other: TimeABC) -> TimeABC:
        self._assert_fps(other)
        return type(self)(self.fps, self.index + other.index)

    def __sub__(self, other: TimeABC) -> TimeABC:
        self._assert_fps(other)
        return type(self)(self.fps, self.index - other.index)

    def __truediv__(self, other: TimeABC) -> float:
        self._assert_fps(other)
        return self.index / other.index

    def _assert_fps(self, other):
        if self.fps != other.fps:
            raise ValueError()

    @property
    def fps(self) -> int:
        """
        Frames per second
        """

        return self._fps

    @property
    def index(self) -> int:
        """
        Time as number of frames
        """

        return self._index

    @property
    def seconds(self) -> float:
        """
        Time in seconds
        """

        return self._index / self._fps # float

    def time(self, index: int) -> TimeABC:
        """
        Generates a new ``Time`` object from a given number of frames based on the time's frames per second.

        Args:
            index : Number of frames
        """

        return type(self)(fps = self._fps, index = index)

    def time_from_seconds(self, seconds: Union[float, int]) -> TimeABC:
        """
        Generates a new ``Time`` object from a given time in seconds based on the time's frames per second.

        Args:
            seconds : Time in seconds
        """

        return type(self).from_seconds(fps = self._fps, seconds = seconds)

    @classmethod
    def from_seconds(cls, fps: int = FPS_DEFAULT, seconds: Union[float, int] = 1.0) -> TimeABC:
        """
        Constructs a new ``Time`` object from frames per second and seconds.

        Args:
            fps : Frames per second
            seconds : Time in seconds
        """

        if isinstance(seconds, int):
            seconds = float(seconds)

        return cls(fps = fps, index = round(seconds * fps))

    @classmethod
    def range(cls, start: TimeABC, stop: TimeABC) -> Generator[TimeABC, None, None]:
        """
        Generator function, similar to Python's range. Generates Time objects instead, frame by frame.

        Args:
            start : Start time of range. ``start.fps`` must be equal to ``stop.fps``.
            stop : Stop time of range (not included). ``start.fps`` must be equal to ``stop.fps``.
        """

        if start.fps != stop.fps:
            raise ValueError()
        if start.index >= stop.index:
            raise ValueError()
        for index in range(start.index, stop.index):
            yield cls(fps = start.fps, index = index)



# -*- coding: utf-8 -*-

"""

BEWEGUNG
a versatile video renderer
https://github.com/pleiszenburg/bewegung

    src/bewegung/core/timescale.py: Time scaling

    Copyright (C) 2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/bewegung/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from datetime import datetime
from typing import Type

from .abc import PyNumber, TimeABC, TimeScaleABC
from .time import Time
from .typeguard import typechecked

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS: TimeScale
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@typechecked
class TimeScale(TimeScaleABC):
    """
    Convert between ``Time`` objects and custom integer & floating point time values.

    Immutable.

    Args:
        start : The time of start in the animation
        start_scaled : A number representing the begin of the the scaled time interval. Must have the same datatype as ``stop_scaled``.
        stop : The time of stop in the animation
        stop_scaled : A number representing the end of the the scaled time interval. Must have the same datatype as ``start_scaled``.
    """

    def __init__(self, start: TimeABC, start_scaled: PyNumber, stop: TimeABC, stop_scaled: PyNumber):

        if start.fps != stop.fps:
            raise ValueError()
        if type(start_scaled) != type(stop_scaled):
            raise TypeError('scaled start and stop times must be of equal type')

        self._start, self._stop, self._start_scaled, self._stop_scaled = start, stop, start_scaled, stop_scaled

        self._dtype = type(start_scaled)
        self._length, self._length_scaled = stop - start, self._stop_scaled - self._start_scaled
        self._scaled2time_factor = self._length.index / self._length_scaled
        self._time2scaled_factor = self._length_scaled / self._length.index

    def __repr__(self) -> str:

        if self._dtype == int:
            return (
                '<TimeScale '
                f'start_scaled={self._start_scaled:d} stop_scaled={self._stop_scaled:d} '
                f'start_index={self._start.index:d} stop_index={self._stop.index:d} fps={self.fps:d} '
                f'dtype={self._dtype.__name__:s}>'
            )

        return (
            '<TimeScale '
            f'start_scaled={self._start_scaled:e} stop_scaled={self._stop_scaled:e} '
            f'start_index={self._start.index:d} stop_index={self._stop.index:d} fps={self.fps:d} '
            f'dtype={self._dtype.__name__:s}>'
        )

    @property
    def dtype(self) -> Type:
        """
        Exposes dtype of object
        """

        return self._dtype

    @property
    def fps(self) -> int:
        """
        Frames per second
        """

        return self._start.fps

    @property
    def start(self) -> TimeABC:
        """
        Start animation time
        """

        return self._start

    @property
    def stop(self) -> TimeABC:
        """
        Stop animation time
        """

        return self._stop

    @property
    def start_scaled(self) -> PyNumber:
        """
        Scaled start time
        """

        return self._start_scaled

    @property
    def stop_scaled(self) -> PyNumber:
        """
        Scaled stop time
        """

        return self._stop_scaled

    def scaled2time(self, scaled_time: PyNumber) -> TimeABC:
        """
        Converts scaled time to time in animation

        Args:
            scaled_time : A number representing scaled time
        """

        if not isinstance(scaled_time, self._dtype):
            raise TypeError('scaled time does not match dtype')

        rel_scaled_time = scaled_time - self._start_scaled
        index = round(rel_scaled_time * self._scaled2time_factor)

        return Time(fps = self._start.fps, index = self._start.index + index)

    def time2scaled(self, time: TimeABC) -> PyNumber:
        """
        Converts time in animation to scaled time

        Args:
            time : Time in animation
        """

        rel_time = time - self._start
        scaled_time = rel_time.index * self._time2scaled_factor

        if self._dtype == int:
            scaled_time = round(scaled_time)

        return self._start_scaled + scaled_time

    @staticmethod
    def dt2msint(dt: datetime) -> int:
        """
        Converts a Python ``datetime.datetime`` object to milliseconds

        Args:
            dt : Date with (optionally) time and (optionally) time zone information attached
        """

        return int(dt.timestamp() * 1000)

    @staticmethod
    def iso2dt(isodate: str) -> datetime:
        """
        Converts a ISO-formatted date to a Python ``datetime.datetime`` object

        Args:
            isodate : A date (and time) following the ISO date format
        """

        return datetime.fromisoformat(isodate)



# -*- coding: utf-8 -*-

"""

BEWEGUNG
a versatile video renderer
https://github.com/pleiszenburg/bewegung

    src/bewegung/core/vector/__init__.py: Vector algebra module root

    Copyright (C) 2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/bewegung/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# EXPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from .single2d import Vector2D
from .single2ddist import Vector2Ddist
from .single3d import Vector3D
from .matrix import Matrix

try:
    import numpy as _np
except ModuleNotFoundError:
    _np = None

if _np is not None:
    from .array2d import VectorArray2D
    from .array2ddist import VectorArray2Ddist
    from .array3d import VectorArray3D

del _np



# -*- coding: utf-8 -*-

"""

BEWEGUNG
a versatile video renderer
https://github.com/pleiszenburg/bewegung

    docs/config.py: Docs config

    Copyright (C) 2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/bewegung/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys

# sys.path.insert(0, os.path.abspath('.'))
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from docs.version import get_version


# -- Project information -----------------------------------------------------

project = 'bewegung'
author = 'Sebastian M. Ernst'
copyright = f'2020 {author:s}'

# The full version, including alpha/beta/rc tags
release = get_version()


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
    'sphinx_autodoc_typehints',
    'sphinx_rtd_theme',
    'sphinxembeddedvideos.youtube', # https://github.com/sphinx-contrib/youtube/issues/9#issuecomment-734295832
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

always_document_param_types = True # sphinx_autodoc_typehints

napoleon_include_special_with_doc = True # napoleon
# napoleon_use_param = True
# napoleon_type_aliases = True



# -*- coding: utf-8 -*-

"""

BEWEGUNG
a versatile video renderer
https://github.com/pleiszenburg/bewegung

    src/bewegung/core/backends/matplotlib.py: Matplotlib backend

    Copyright (C) 2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/bewegung/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from typing import Any, Callable
import warnings

from PIL.Image import Image, fromarray, frombuffer, merge

from ._base import BackendBase
from ..abc import ColorABC, NumberTypes, VideoABC
from ..typeguard import typechecked

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@typechecked
class Backend(BackendBase):

    _name = 'Matplotlib'

    def __init__(self):

        super().__init__()

        self._plt, self._Figure = None, None

        self._mplcairo_present = False

    def _prototype(self, video: VideoABC, **kwargs) -> Callable:

        if 'dpi' not in kwargs.keys():
            kwargs['dpi'] = 300 # important!
        if 'figsize' not in kwargs.keys():
            if 'width' not in kwargs.keys():
                kwargs['width'] = video.width
            if 'height' not in kwargs.keys():
                kwargs['height'] = video.height
            assert isinstance(kwargs['width'], NumberTypes)
            assert isinstance(kwargs['height'], NumberTypes)
            kwargs['figsize'] = (
                kwargs.pop('width') / kwargs['dpi'],
                kwargs.pop('height') / kwargs['dpi'],
            ) # inch

        if 'background_color' in kwargs.keys() and 'facecolor' in kwargs.keys():
            kwargs.pop('background_color')
        if 'background_color' in kwargs.keys():
            if not isinstance(kwargs['background_color'], ColorABC):
                raise TypeError('color expected')
            kwargs['facecolor'] = f'#{kwargs.pop("background_color").as_hex():s}'
        if 'facecolor' not in kwargs.keys():
            kwargs['facecolor'] = '#FFFFFF00'

        if 'tight_layout' not in kwargs.keys():
            kwargs['tight_layout'] = True

        managed = kwargs.pop('managed') if 'managed' in kwargs.keys() else True
        assert isinstance(managed, bool)

        @typechecked
        def new_figure() -> self._Figure:
            fig = self._type(**kwargs)
            if managed:
                setattr(fig, '__bewegung_managed__', None) # flag: close figure after extracting image
            return fig

        return new_figure

    def _isinstance(self, obj: Any) -> bool:

        return isinstance(obj, self._Figure) # Return type is a Figure object!

    def _load(self):

        try:
            import mplcairo.base # import before matplotlib
            self._mplcairo_present = True
        except ModuleNotFoundError:
            warnings.warn('`mplcairo` matplotlib backend is not installed, falling back to `cairo`')

        import matplotlib
        matplotlib.use(
            "module://mplcairo.base" if self._mplcairo_present else "cairo",
            force = True,
        ) # use mplcairo.base as non-GUI backend

        import matplotlib.pyplot as plt # import pyplot last
        from matplotlib.figure import Figure

        self._type = plt.figure

        self._plt = plt
        self._Figure = Figure

    def _to_pil(self, obj: Any) -> Image:

        if self._mplcairo_present:

            obj.canvas.draw()

            buffer = obj.canvas.renderer.buffer_rgba()
            assert buffer.dtype.name == 'uint8' # TODO cairo & mplcairo also support RGBA128F

            image = fromarray(buffer) # depends on matplotlib backend - https://stackoverflow.com/q/57316491/1672565

        else:

            surface = obj.canvas._get_printed_image_surface() # returns ARGB32 cairo surface

            image = frombuffer(
                mode = 'RGBa',
                size = (surface.get_width(), surface.get_height()),
                data = surface.get_data().tobytes(), # call to "tobytes" required because of RGBa mode
                )
            b, g, r, a = image.split()
            image = merge('RGBa', (r, g, b, a)).convert("RGBA")

        if hasattr(obj, '__bewegung_managed__'): # close flagged image
            self._plt.close(obj)

        assert image.mode == 'RGBA'

        return image



# -*- coding: utf-8 -*-

"""

BEWEGUNG
a versatile video renderer
https://github.com/pleiszenburg/bewegung

    src/bewegung/core/backends/pillow.py: Pillow backend

    Copyright (C) 2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/bewegung/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from typing import Any, Callable

from PIL.Image import Image, new

from ._base import BackendBase
from ..abc import ColorABC, VideoABC
from ..typeguard import typechecked

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@typechecked
class Backend(BackendBase):

    _name = 'Pillow'

    def _prototype(self, video: VideoABC, **kwargs) -> Callable:

        if 'mode' not in kwargs.keys():
            kwargs['mode'] = 'RGBA'

        if 'size' in kwargs.keys() and 'width' in kwargs.keys():
            kwargs.pop('width')
        if 'size' in kwargs.keys() and 'height' in kwargs.keys():
            kwargs.pop('height')
        if 'size' not in kwargs.keys():
            kwargs['size'] = (video.width, video.height)
        else:
            if 'width' not in kwargs.keys() and 'height' not in kwargs.keys():
                raise ValueError('width or height missing')
            kwargs['size'] = (kwargs.pop('width'), kwargs.pop('height'))

        if 'color' in kwargs.keys() and 'background_color' in kwargs.keys():
            kwargs.pop('background_color')
        if 'background_color' in kwargs.keys():
            if not isinstance(kwargs['background_color'], ColorABC):
                raise TypeError('color expected')
            kwargs['color'] = kwargs.pop("background_color").as_rgba_int()

        return lambda: new(**kwargs)

    def _load(self):

        self._type = Image

    def _to_pil(self, obj: Any) -> Image:

        if obj.mode != 'RGBA':
            raise TypeError('unhandled image mode')

        return obj



# -*- coding: utf-8 -*-

"""

BEWEGUNG
a versatile video renderer
https://github.com/pleiszenburg/bewegung

    src/bewegung/core/camera.py: Simple pinhole camera

    Copyright (C) 2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/bewegung/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import math
import sys
from typing import Union

try:
    import numpy as np
except ModuleNotFoundError:
    np = None

try:
    from numba import jit, float32, float64, boolean
except ModuleNotFoundError:
    def jit(*jit_args, **jit_kwargs):
        def wrapper(func):
            return func
        return wrapper
    boolean, float32, float64 = None, tuple(), tuple()

from .abc import CameraABC
from .typeguard import typechecked
from .vector import (
    Matrix,
    Vector2D,
    Vector2Ddist,
    Vector3D,
    VectorArray2Ddist,
    VectorArray3D,
    )

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@typechecked
class Camera(CameraABC):
    """
    A virtual camera for 3D to 2D projections.
    3D vectors are projected onto a 2D plane and returned combined with the absolute distance to the camera in 3D space.
    Vector arrays can be handled if ``numpy`` and ``numba`` are present.

    Mutable.

    Args:
        position : Position in 3D space
        direction : Direction of view in 3D space
        roll : Roll angle in radians
        planeOffset : Center offset of 2D plane
        planeFactor : Scaling factor for the 2D plane
        planeYFlip : Allows to invert y-axes of 2D plane, i.e. switch to positive downwards
    """

    def __init__(self,
        position: Union[Vector3D, None] = None,
        direction: Union[Vector3D, None] = None,
        roll: float = 0.0,
        planeOffset: Union[Vector2D, None] = None,
        planeFactor: float = 1.0,
        planeYFlip: bool = False,
    ):

        # camera position
        self._position = Vector3D(0.0, 0.0, 0.0) if position is None else position
        # 3D view vector
        self._direction = Vector3D(1.0, 0.0, 0.0) if direction is None else direction
        if not math.isclose(self._direction.mag, 1.0):
            self._direction.mul(1.0 / self._direction.mag)
        self._roll = roll

        # Center offset on 2D rendering plane
        self._planeOffset = Vector2D(0.0, 0.0) if planeOffset is None else planeOffset
        # plane scale factor
        self._planeFactor = planeFactor
        # place Y flip
        self._planeYFlip = planeYFlip

        # 2D rendering plane in 3D space
        self._planeX = Vector3D(0.0, 0.0, 0.0)
        self._planeY = Vector3D(0.0, 0.0, 0.0)
        # compute rendering plane
        self._update_plane()

    def __repr__(self) -> str:

        return (
            '<Camera '
            f'id={id(self):x} '
            f'x={self._position.x:e} y={self._position.y:e} z={self._position.z:e} '
            f'dx={self._direction.x:e} dy={self._direction.y:e} dz={self._direction.z:e}>'
            )

    def _update_plane(self):

        _, theta, phi = self._direction.as_polar_tuple()
        theta = (math.pi / 2) - theta

        directionXYmag = Vector2D(
            x = self._direction.x,
            y = self._direction.y,
            ).mag
        SinTheta = math.sin(theta)
        tmp = 1.0 / Vector3D(
            x = self._direction.x * SinTheta,
            y = self._direction.y * SinTheta,
            z = -directionXYmag,
            ).mag

        self._planeX.update(
            x = math.sin(phi),
            y = -math.cos(phi),
            z = 0.0,
            )
        self._planeY.update(
            x = self._direction.x * SinTheta * tmp,
            y = self._direction.y * SinTheta * tmp,
            z = -directionXYmag * tmp,
            )

        assert math.isclose(self._planeX.mag, 1.0)
        assert math.isclose(self._planeY.mag, 1.0)

        if self._roll == 0:
            return

        R = Matrix.from_3d_rotation(v = self._direction, a = self._roll)
        self._planeX = R @ self._planeX
        self._planeY = R @ self._planeY

    @property
    def direction(self) -> Vector3D:
        """
        Direction of view in 3D space
        """

        return self._direction

    @direction.setter
    def direction(self, value: Vector3D):
        """
        Direction of view in 3D space
        """

        self._direction.update_from_vector(value)
        if not math.isclose(self._direction.mag, 1.0):
            self._direction.mul(1.0 / self._direction.mag)
        self._update_plane()

    @property
    def roll(self) -> float:
        """
        Roll angle in radians
        """

        return self._roll

    @roll.setter
    def roll(self, value: float):
        """
        Roll angle in radians
        """

        self._roll = value
        self._update_plane()

    @property
    def position(self) -> Vector3D:
        """
        Position in 3D space
        """

        return self._position

    @position.setter
    def position(self, value: Vector3D):
        """
        Position in 3D space
        """

        self._position.update_from_vector(value)

    @property
    def planeFactor(self) -> float:
        """
        Scaling factor for the 2D plane
        """

        return self._planeFactor

    @planeFactor.setter
    def planeFactor(self, value: float):
        """
        Scaling factor for the 2D plane
        """

        self._planeFactor = value

    @property
    def planeOffset(self) -> Vector2D:
        """
        Center offset of 2D plane
        """

        return self._planeOffset

    @planeOffset.setter
    def planeOffset(self, value: Vector2D):
        """
        Center offset of 2D plane
        """

        self._planeOffset.update_from_vector(value)

    @property
    def planeYFlip(self) -> bool:
        """
        Allows to invert y-axes of 2D plane, i.e. switch to positive downwards
        """

        return self._planeYFlip

    @planeYFlip.setter
    def planeYFlip(self, value: bool):
        """
        Allows to invert y-axes of 2D plane, i.e. switch to positive downwards
        """

        self._planeYFlip = value

    def get_point(self, point3D: Vector3D) -> Vector2Ddist:
        """
        Projects a 3D vector onto a 2D plane.
        Returns a 2D vector combined with the absolute distance to the camera in 3D space.

        Args:
            point3D : point in 3D space
        """

        ma = [
            [self._planeX.x, self._planeY.x, -(point3D.x - self._position.x), -self._direction.x],
            [self._planeX.y, self._planeY.y, -(point3D.y - self._position.y), -self._direction.y],
            [self._planeX.z, self._planeY.z, -(point3D.z - self._position.z), -self._direction.z],
            ]

        determ = (
              ma[0][0] * ma[1][1] * ma[2][2]
            + ma[0][1] * ma[1][2] * ma[2][0]
            + ma[0][2] * ma[1][0] * ma[2][1]
            - ma[0][2] * ma[1][1] * ma[2][0]
            - ma[0][0] * ma[1][2] * ma[2][1]
            - ma[0][1] * ma[1][0] * ma[2][2]
            )

        if determ == 0.0:
            determ = sys.float_info.min # HACK

        point2D = Vector2D(
            x = (
                  ma[0][3] * ma[1][1] * ma[2][2]
                + ma[0][1] * ma[1][2] * ma[2][3]
                + ma[0][2] * ma[1][3] * ma[2][1]
                - ma[0][2] * ma[1][1] * ma[2][3]
                - ma[0][3] * ma[1][2] * ma[2][1]
                - ma[0][1] * ma[1][3] * ma[2][2]
                ) / determ,
            y = (
                  ma[0][0] * ma[1][3] * ma[2][2]
                + ma[0][3] * ma[1][2] * ma[2][0]
                + ma[0][2] * ma[1][0] * ma[2][3]
                - ma[0][2] * ma[1][3] * ma[2][0]
                - ma[0][0] * ma[1][2] * ma[2][3]
                - ma[0][3] * ma[1][0] * ma[2][2]
                ) / determ,
            )

        if self._planeYFlip:
            point2D.y = -point2D.y

        point2D.x *= self._planeFactor
        point2D.y *= self._planeFactor
        point2D += self._planeOffset

        return Vector2Ddist(
            x = point2D.x,
            y = point2D.y,
            dist = (point3D - self._position).mag,
            )

    def get_points(self, points3d: VectorArray3D) -> VectorArray2Ddist:
        """
        Projects a 3D vector array onto a 2D plane.
        Returns a 2D vector array combined with the absolute distances to the camera in 3D space.

        Args:
            points3d : points in 3D space
        """

        if np is None:
            raise NotImplementedError('numpy is not available')

        position = self._position.as_ndarray() # type
        planeOffset = self._planeOffset.as_ndarray() # type
        points3d = points3d.as_ndarray() # type
        planeFactor = np.float32(self._planeFactor) # type

        ma = np.array([
            [self._planeX.x, self._planeY.x, 0.0, -self._direction.x],
            [self._planeX.y, self._planeY.y, 0.0, -self._direction.y],
            [self._planeX.z, self._planeY.z, 0.0, -self._direction.z],
            ], dtype = points3d.dtype) # type / matrix
        empty = np.array([
            np.nan,
            np.nan,
            np.nan,
            ], dtype = points3d.dtype) # NaN placeholder
        points2d = np.zeros(points3d.shape, dtype = points3d.dtype) # type / target

        self._get_points_jit(
            points3d, points2d,
            ma, position, empty, planeOffset,
            planeFactor, self._planeYFlip,
            )

        return VectorArray2Ddist(
            x = points2d[:, 0],
            y = points2d[:, 1],
            dist = points2d[:, 2],
            )

    @staticmethod
    @jit(
        [
            (
                float32[:, :], float32[:, :],
                float32[:, :], float32[:], float32[:], float32[:],
                float32, boolean,
            ), (
                float64[:, :], float64[:, :],
                float64[:, :], float64[:], float64[:], float64[:],
                float64, boolean,
            )
        ],
        nopython = True,
    )
    def _get_points_jit(
        points_3d, points_2d,
        ma, position, empty, planeOffset,
        planeFactor, planeYFlip,
        ):

        for index in range(0, points_3d.shape[0]):

            ma[:, 2] = position - points_3d[index, :]

            determ = (
                  ma[0][0] * ma[1][1] * ma[2][2]
                + ma[0][1] * ma[1][2] * ma[2][0]
                + ma[0][2] * ma[1][0] * ma[2][1]
                - ma[0][2] * ma[1][1] * ma[2][0]
                - ma[0][0] * ma[1][2] * ma[2][1]
                - ma[0][1] * ma[1][0] * ma[2][2]
                )

            if determ == 0:
                points_2d[index, :] = empty
                continue

            points_2d[index, 0] = (
                  ma[0][3] * ma[1][1] * ma[2][2]
                + ma[0][1] * ma[1][2] * ma[2][3]
                + ma[0][2] * ma[1][3] * ma[2][1]
                - ma[0][2] * ma[1][1] * ma[2][3]
                - ma[0][3] * ma[1][2] * ma[2][1]
                - ma[0][1] * ma[1][3] * ma[2][2]
                )
            points_2d[index, 1] = (
                  ma[0][0] * ma[1][3] * ma[2][2]
                + ma[0][3] * ma[1][2] * ma[2][0]
                + ma[0][2] * ma[1][0] * ma[2][3]
                - ma[0][2] * ma[1][3] * ma[2][0]
                - ma[0][0] * ma[1][2] * ma[2][3]
                - ma[0][3] * ma[1][0] * ma[2][2]
                )

            points_2d[index, :2] *= planeFactor / determ
            if planeYFlip:
                points_2d[index, 1] = -points_2d[index, 1]
            points_2d[index, :2] += planeOffset

            points_2d[index, 2] = np.sqrt(np.sum(np.power(points_3d[index, :] - position, 2))) # distance



# -*- coding: utf-8 -*-

"""

BEWEGUNG
a versatile video renderer
https://github.com/pleiszenburg/bewegung

    src/bewegung/core/indexpool.py: Pools of unique index values

    Copyright (C) 2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/bewegung/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from typing import List

from .abc import IndexPoolABC
from .typeguard import typechecked

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@typechecked
class IndexPool(IndexPoolABC):
    """
    Used to manage z-index for layers tasks and prepare-order ("preporder") values ("indices") for prepare tasks.
    Indices, represented by integers, are unique within a pool.
    New pools are empty.

    Mutable.
    """

    def __init__(self):

        self._pool = set()

    def __repr__(self) -> str:

        return f'<IndexPool len={len(self):d}>'

    def __contains__(self, index: int) -> bool:
        """
        Checks whether an index is already present in the pool.

        Args:
            index : Value that should be checked
        """

        return index in self._pool

    def __len__(self) -> int:
        """
        Number of entries in the pool
        """

        return len(self._pool)

    @property
    def max(self) -> int:
        """
        Highest index currently present in pool
        """

        return max(self._pool)

    @property
    def min(self) -> int:
        """
        Lowest index currently present in pool
        """

        return min(self._pool)

    def as_list(self) -> List[int]:
        """
        Exports pool as a sorted list
        """

        return sorted(self._pool)

    def on_top(self) -> int:
        """
        Returns an index that would be on top of the pool, i.e. ``max + 1``. If the pool is empty, it returns 0.
        The index is not automatically added (registered) to the pool.
        """

        return (self.max + 1) if len(self) != 0 else 0

    def on_bottom(self) -> int:
        """
        Returns an index that would be on the bottom of the pool, i.e. ``min - 1``. If the pool is empty, it returns 0.
        The index is not automatically added (registered) to the pool.
        """

        return (self.min - 1) if len(self) != 0 else 0

    def register(self, index: int):
        """
        Registers, i.e. adds an index to the pool. Raises an exception if the index is already present in the pool.

        Args:
            index : Value that should be added to the pool
        """

        if index in self:
            raise IndexError('index is already present in pool')

        self._pool.add(index)



# -*- coding: utf-8 -*-

"""

BEWEGUNG
a versatile video renderer
https://github.com/pleiszenburg/bewegung

    src/bewegung/core/vector/array2ddist.py: 2D Vector Array with distance parameter

    Copyright (C) 2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/bewegung/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from typing import List, Union

import numpy as np
from typeguard import typechecked

from .lib import dtype_np2py
from .single2ddist import Vector2Ddist
from .array2d import VectorArray2D
from ..abc import Dtype, Number, Vector2DABC, VectorArray2DABC, VectorIterable2D
from ..const import FLOAT_DEFAULT

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@typechecked
class VectorArray2Ddist(VectorArray2D):
    """
    Immutable
    """

    def __init__(self, x: np.ndarray, y: np.ndarray, dist: np.ndarray):
        assert dist.ndim == 1
        assert x.shape[0] == y.shape[0] == dist.shape[0]
        assert x.dtype == y.dtype == dist.dtype
        x.setflags(write = False)
        y.setflags(write = False)
        dist.setflags(write = False)
        super().__init__(x = x, y = y,)
        self._dist = dist

    def __repr__(self) -> str:
        return f'<VectorArray2Ddist len={len(self):d}>'

    def __getitem__(self, idx: Union[int, slice]) -> Union[Vector2Ddist, VectorArray2DABC]:
        if isinstance(idx, int):
            dtype = dtype_np2py(self.dtype)
            return Vector2Ddist(dtype(self._x[idx]), dtype(self._y[idx]), dtype(self._dist[idx]), dtype = dtype)
        return VectorArray2Ddist(self._x[idx].copy(), self._y[idx].copy(), self._dist[idx].copy())

    def mul(self, scalar: Number):
        raise NotImplementedError()

    def as_vectorarray(self) -> VectorArray2DABC:
        return VectorArray2D(self._x.copy(), self._y.copy())

    def as_list(self) -> List[Vector2DABC]:
        dtype = dtype_np2py(self.dtype)
        return [
            Vector2Ddist(dtype(self._x[idx]), dtype(self._y[idx]), dtype(self._dist[idx]), dtype = dtype)
            for idx in range(len(self))
        ]

    def as_ndarray(self, dtype: Dtype = FLOAT_DEFAULT) -> np.ndarray:
        a = np.zeros((len(self), 3), dtype = self.dtype)
        a[:, 0], a[:, 1], a[:, 2] = self._x, self._y, self._dist
        return a if a.dtype == np.dtype(dtype) else a.astype(dtype)

    def as_type(self, dtype: Dtype) -> VectorArray2DABC:
        return self.copy() if self.dtype == np.dtype(dtype) else VectorArray2Ddist(
            self._x.astype(dtype), self._y.astype(dtype), self._dist.astype(dtype),
        )

    def copy(self) -> VectorArray2DABC:
        return VectorArray2Ddist(self._x.copy(), self._y.copy(), self._dist.copy())

    def update_from_vector(self, other: VectorArray2DABC):
        raise NotImplementedError()

    @property
    def dist(self) -> np.ndarray:
        return self._dist
    @dist.setter
    def dist(self, value: float):
        raise NotImplementedError()

    @classmethod
    def from_iterable(cls, obj: VectorIterable2D, dtype: Dtype = FLOAT_DEFAULT) -> VectorArray2DABC:
        if not isinstance(obj, list):
            obj = list(obj)
        assert all((isinstance(item, Vector2Ddist) for item in obj))
        x = np.zeros((len(obj),), dtype = dtype)
        y = np.zeros((len(obj),), dtype = dtype)
        dist = np.zeros((len(obj),), dtype = dtype)
        for idx, item in enumerate(obj):
            x[idx], y[idx], dist[idx] = item.x, item.y, item.dist
        return cls(x = x, y = y, dist = dist,)

    @classmethod
    def from_polar(cls, radius: np.ndarray, angle: np.ndarray) -> VectorArray2DABC:
        raise NotImplementedError()



# -*- coding: utf-8 -*-

"""

BEWEGUNG
a versatile video renderer
https://github.com/pleiszenburg/bewegung

    src/bewegung/core/video.py: Parallel video frame renderer

    Copyright (C) 2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/bewegung/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import inspect
import multiprocessing as mp
from typing import Callable, Dict, Union

from PIL import Image as PIL_Image
try:
    from tqdm import tqdm
except ModuleNotFoundError:
    tqdm = lambda x: x

from .abc import EncoderABC, LayerABC, SequenceABC, VideoABC, Vector2DABC, TimeABC
from .backends import backends
from .const import FPS_DEFAULT
from .encoders import FFmpegH264Encoder
from .indexpool import IndexPool
from .layer import Layer
from .sequence import Sequence
from .task import Task
from .time import Time
from .typeguard import typechecked
from .vector import Vector2D

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# "GLOBALS" (FOR WORKERS)
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

_workers = {}

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@typechecked
class Video(VideoABC):
    """
    This class is the "core" of ``bewegung``.
    It generates video objects.
    It manages sequences, layers and preparation tasks and is responsible for rendering the actual video file.

    Video objects are mutable.

    Args:
        width : Video width in pixels
        height : Video width in pixels
        ctx : Context, store for any type of video context data
        fps : Frames per second
        seconds : Duration of video in seconds (``frames`` and ``seconds`` are mutually exclusive. Specify excactly one of two.)
        frames : Duration of video as number of frames (``frames`` and ``seconds`` are mutually exclusive. Specify excactly one of two.)
    """

    def __init__(self,
        width: int,
        height: int,
        ctx: Union[Dict, None] = None,
        fps: int = FPS_DEFAULT,
        seconds: Union[float, int, None] = None,
        frames: Union[int, None] = None,
    ):

        if width <= 0:
            raise ValueError('width must be greater than 0')
        if height <= 0:
            raise ValueError('height must be greater than 0')

        self._width = width
        self._height = height
        self._ctx = ctx if ctx is not None else {}

        if fps <= 0:
            raise ValueError('fps must be greater than 0')
        if not ((seconds is not None) ^ (frames is not None)):
            raise ValueError('frames and seconds are mutually exclusive')
        if seconds is not None:
            if seconds <= 0:
                raise ValueError('seconds must be greater than 0')
        if frames is not None:
            if frames <= 0:
                raise ValueError('frames must be greater than 0')
        self._length = Time(
            fps = fps, index = frames,
            ) if seconds is None else Time.from_seconds(
            fps = fps, seconds = seconds,
        )

        self._sequences = [] # list of sequences

        self._preptasks = [] # list of sequence prepare tasks
        self._layertasks = [] # list of layer render tasks
        self._preporder = IndexPool()
        self._zindex = IndexPool()

    def __repr__(self) -> str:
        """
        String representation for interactive use
        """

        return f'<Video frames={self._length.index:d} seconds={self._length.seconds:.03f}s fps={self.fps:d}>'

    def __len__(self) -> int:
        """
        Duration of video as number of frames
        """

        return self._length.index

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# PROPERTIES
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    @property
    def length(self) -> TimeABC:
        """
        Duration of video
        """

        return self._length

    @property
    def fps(self) -> int:
        """
        Frames per second
        """

        return self._length.fps

    @property
    def width(self) -> int:
        """
        Width of video in pixels
        """

        return self._width

    @property
    def height(self) -> int:
        """
        Height of video in pixels
        """

        return self._height

    @property
    def ctx(self) -> Dict:
        """
        Context (mutable), store for any type of video context data
        """

        return self._ctx

    @property
    def preporder(self) -> IndexPool:
        """
        Prepare-order index pool for prepare tasks (mutable)
        """

        return self._preporder

    @property
    def zindex(self) -> IndexPool:
        """
        Z-index pool for layers (mutable)
        """

        return self._zindex

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# RESET TASKS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    def time(self, index: int) -> TimeABC:
        """
        Generates a new :class:`bewegung.Time` object from a given number of frames based on the video's frames per second.

        Args:
            index : Number of frames
        """

        return self._length.time(index = index)

    def time_from_seconds(self, seconds: Union[float, int]) -> TimeABC:
        """
        Generates a new :class:`bewegung.Time` object from a given time in seconds based on the video's frames per second.

        Args:
            seconds : Time in seconds
        """

        return self._length.time_from_seconds(seconds = seconds)

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# RESET TASKS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    def reset(self):
        """
        This method allows to reset a video object in preparation of a new render run.
        It is automatically invoked when calling ``Video.render``.
        It may instead be used before rendering indivual frames with ``Video.render_frame``.
        """

        for sequence in self._sequences:
            sequence.reset()

        self._preptasks.clear()
        self._preptasks.extend([
            Task(
                sequence = sequence,
                index = getattr(sequence, attr).preporder_tag,
                task = getattr(sequence, attr),
            )
            for sequence in self._sequences for attr in dir(sequence)
            if hasattr(getattr(sequence, attr), 'preporder_tag')
        ]) # find prepare methods based on tags
        self._preptasks.sort() # sort by preporder

        self._layertasks.clear()
        self._layertasks.extend([
            Task(
                sequence = sequence,
                index = getattr(sequence, attr).zindex_tag,
                task = getattr(sequence, attr),
            )
            for sequence in self._sequences for attr in dir(sequence)
            if hasattr(getattr(sequence, attr), 'zindex_tag')
        ]) # find layer methods based on tags
        self._layertasks.sort() # sort by (z-) index

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# DECORATOR: SEQUENCE (TYPE)
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    def sequence(self,
        start: Union[Time, None] = None,
        stop: Union[Time, None] = None,
        length: Union[Time, None] = None,
    ) -> Callable:
        """
        A **decorator** for decorating user-defined ``sequence`` classes.
        New classes are created by making :class:`bewegung.core.sequence.Sequence` inherit from the user-defined sequence classes.

        Args:
            start : Time of start of sequence within the video.
                A negative time can be used to specify a time relative to the end of the video.
                Defaults to the beginning of the video.
            stop : Time of stop of sequence within the video.
                A negative time can be used to specify a time relative to the end of the video.
                Defaults to the end of the video.
                ``stop`` and ``length`` are mutually exclusive. Specify one at most.
            length : Length of the sequence.
                ``stop`` and ``length`` are mutually exclusive. Specify one at most.
        """

        if start is None:
            start = self.time(0)
        if start.fps != self.fps:
            start = self.time_from_seconds(seconds = start.seconds)

        if stop is not None and length is not None:
            raise ValueError('stop and length are mutually exclusive')
        if length is not None:
            if length.fps != length.fps:
                length = self.time_from_seconds(seconds = length.seconds)
            stop = length - start
        if stop is not None:
            if stop.fps != self.fps:
                stop = self.time_from_seconds(seconds = stop.seconds)
        else:
            stop = self._length

        if start < self.time(0):
            start = self._length + start
        if not (self.time(0) <= start <= self._length):
            raise ValueError('start out of bounds')
        if stop < self.time(0):
            stop = self._length + stop
        if not (self.time(0) <= stop <= self._length):
            raise ValueError('stop out of bounds')

        if start >= stop:
            raise ValueError('stop smaller or equal to stop')

        @typechecked
        def decorator(cls: type):

            cls_bases, sequence_bases = inspect.getmro(cls), inspect.getmro(Sequence)
            bases = tuple([item for item in sequence_bases if item not in cls_bases]) + cls_bases
            SequenceCls = type(cls.__name__, bases, Sequence.__dict__.copy())
            sequence = SequenceCls(start = start, stop = stop, video = self)

            self._sequences.append(sequence)
            return sequence # HACK return object, not class

        return decorator

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# DECORATOR: PREPARE (TASK)
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    def prepare(self,
        preporder: Union[int, None] = None,
    ) -> Callable:
        """
        A **decorator** for decorating ``prepare`` methods (tasks) within ``sequence`` classes.

        Args:
            preporder : A number, managed by an index pool, representing the relative position within a set of ``prepare`` tasks. If not provided, the new task will be created at the end of the set.
        """

        if preporder is None:
            preporder = self._preporder.on_top()

        self._preporder.register(preporder) # ensure unique preporder

        @typechecked
        def decorator(method: Callable) -> Callable:

            @typechecked
            def wrapper(sequence: SequenceABC, time: Time):

                kwargs = {}
                for param in method.__code__.co_varnames[
                    1:method.__code__.co_argcount # excluding self and internal namespace
                ]: # parameters requested by user
                    if param == 'time':
                        kwargs[param] = time
                    elif param == 'reltime':
                        kwargs[param] = time - sequence.start
                    else:
                        raise ValueError('unknown parameter', param)

                method(sequence, **kwargs) # let user draw layer/canvas for frame

            wrapper.preporder_tag = preporder # tag wrapper function
            return wrapper

        return decorator

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# DECORATOR: LAYER (TASK)
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    def layer(self,
        zindex: Union[int, None] = None,
        canvas: Union[Callable, None] = None,
        offset: Union[Vector2DABC, None] = None,
    ) -> Callable:
        """
        A **decorator** for decorating ``layer`` methods (tasks) within ``sequence`` classes.
        Methods are wrapped with callable :class:`bewegung.core.layer.Layer` objects.

        Args:
            zindex : A number, managed by an index pool, representing the relative position within a stack of ``layer`` tasks. If not provided, the new layer will be created on top.
            canvas : A function pointer, generating a new canvas once per frame for the ``layer`` task.
            offset : The layer's offset relative to the top-left corner of the video. The y-axis is downwards positive.
        """

        if offset is None:
            offset = Vector2D(0, 0)
        if zindex is None:
            zindex = self._zindex.on_top()

        self._zindex.register(zindex) # ensure unique z-index

        @typechecked
        def decorator(method: Callable) -> LayerABC:

            return Layer(
                method = method,
                zindex = zindex,
                video = self,
                canvas = canvas,
                offset = offset,
            ) # callable object (pretending to be a method)

        return decorator

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CANVAS PROTOTYPES FOR LAYER
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    def canvas(self, backend: str = 'drawingboard', **kwargs) -> Callable:
        """
        A method to create function pointers for functions generating new canvases.
        The pointers can be passed to the ``canvas`` parameter in the ``layer`` decorator method.

        Args:
            backend : Selected type of canvas, i.e. name of desired backend.
            kwargs : Keyword arguments of the selected backend's canvas creation function.
        """

        return backends[backend].prototype(video = self, **kwargs)

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# RENDER VIDEO
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    def render(self,
        processes: int = 1,
        batchsize: int = 256,
        encoder: Union[EncoderABC, None] = None,
        frame_fn: Union[str, None] = None,
        video_fn: Union[str, None] = None,
        ):
        """
        A method for rendering the actual video file.

        This function starts at least one Python sub-process (worker process) for rendering frames.
        Based on multiple worker processes, multiple frames can be rendered in parallel.
        If a filename for a video is specified, the rendered frames are streamed to a video encoder.

        Args:
            processes : Number of parallel frame rendering (worker) processes
            batchsize : Maximum number of frames rendered by a worker process before the (old) worker is replaced by a new worker.
                This option helps to prevent long rendering jobs from running out of memory.
            encoder : A video encoder object.
                If omitted, the ffmpeg H.264 encoder will be selected.
            frame_fn : A Python string template (representing a path) including an ``index`` integer placeholder.
                If specified, individual frames will be stored here.
                If omitted, no frames will be stored.
            video_fn: Location and name (path) of where to store the video file.
                If omitted, no video will be rendered.
                However, indivual frames may in fact still be rendered if ``frame_fn`` has been specified.
        """

        if processes <= 0:
            raise ValueError('processes must be greater than 0')
        if batchsize <= 0:
            raise ValueError('batchsize must be greater than 0')

        if video_fn is not None and len(video_fn) == 0:
            raise ValueError('if a string, video_fn must not be empty')
        if video_fn is not None and encoder is None:
            encoder = FFmpegH264Encoder()

        self.reset()

        workers = mp.Pool(
            processes = processes,
            initializer = self._worker_init,
            initargs = (self,),
            maxtasksperchild = batchsize,
        )
        workers_promises = [
            workers.apply_async(
                func = self._worker_render_frame,
                args = (time, video_fn is not None, frame_fn),
                error_callback = self._worker_error,
            ) for time in Time.range(self.time(0), self._length)
        ]

        if video_fn is None:

            for promise in tqdm(workers_promises):
                _ = promise.get()

        else:

            with encoder(video = self, video_fn = video_fn) as stream:
                for promise in tqdm(workers_promises):
                    frame = promise.get()
                    frame.save(stream, 'bmp')
                    stream.flush()
                    frame.close()

        workers.close()
        workers.terminate()
        workers.join()

    def render_frame(self,
        time: Time,
        return_frame: bool = True,
        frame_fn: Union[str, None] = None,
        ) -> Union[PIL_Image.Image, None]:
        """
        A method for rendering individual frames of the video.

        The frames can optionally be written to disk and/or returned to the caller.
        **Before calling this method for the first time, ``Video.reset`` has to be invoked!**

        Args:
            time : Time of the frame relative to the beginning of the video
            return_frame : If ``True``, the frame is returned to the caller of the method.
            frame_fn: Location and name (path) of where to store the rendered frame.
                If omitted, the frame is not stored.
        Returns:
            If requested via ``return_frame``, a pillow image object is returned.
        """

        for preptask in self._preptasks:
            if time in preptask.sequence:
                preptask(time)

        layers = [
            layertask(time)
            for layertask in self._layertasks
            if time in layertask.sequence # only render layer if time within sequence
        ] # call layer render functions, get list of uni-size PIL images

        base_layer = PIL_Image.new('RGBA', (self._width, self._height), (0, 0, 0, 0)) # transparent black
        for layer in layers:
            base_layer.paste(im = layer, box = layer.offset.as_tuple(), mask = layer)

        base_layer = base_layer.convert('RGB') # go from RGBA to RGB

        if frame_fn is not None:
            base_layer.save(frame_fn.format(index = time.index))

        if return_frame:
            return base_layer # for direct to video

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# WORKER INFRASTRUCTURE
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    @staticmethod
    def _worker_error(err: Exception):

        raise err

    @staticmethod
    def _worker_init(video: VideoABC):

        _workers[mp.current_process().name] = video

    @staticmethod
    def _worker_render_frame(*args, **kwargs): # transparent wrapper for `render_frame`

        return _workers[mp.current_process().name].render_frame(*args, **kwargs)



# -*- coding: utf-8 -*-

"""

BEWEGUNG
a versatile video renderer
https://github.com/pleiszenburg/bewegung

    src/bewegung/core/vector/matrix.py: Simple 2x2/3x3 matrix for rotations

    Copyright (C) 2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/bewegung/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from math import cos, sin
from typing import List, Tuple, Type, Union

try:
    import numpy as np
    from numpy import ndarray
except ModuleNotFoundError:
    np, ndarray = None, None
from typeguard import typechecked

from ..abc import Dtype, MatrixABC, PyNumber, Vector2DABC, Vector3DABC
from ..const import FLOAT_DEFAULT
from .single2d import Vector2D
from .single3d import Vector3D

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@typechecked
class Matrix(MatrixABC):
    """
    Mutable
    """

    def __init__(self, matrix = List[List[PyNumber]], dtype: Union[Type, None] = None):

        lines = len(matrix)
        assert lines in (2, 3) # allow 2D and 3D
        assert all((len(line) == lines for line in matrix))

        if dtype is None:
            dtype = type(matrix[0][0])
        else:
            assert isinstance(matrix[0][0], dtype)
        assert all((all((isinstance(number, dtype) for number in line)) for line in matrix))

        self._dtype = dtype
        self._matrix = matrix

    def __repr__(self) -> str:
        return f'<Matrix shape={len(self._matrix):d}x{len(self._matrix):d} dtype={self._dtype.__name__:s}>'

    def __matmul__(self, vector: Union[Vector2DABC, Vector3DABC]) -> Union[Vector2DABC, Vector3DABC]:

        vector = vector.as_tuple()
        assert len(self._matrix) == len(vector)

        values = [
            sum([item_a * item_b for item_a, item_b in zip(line, vector)])
            for line in self._matrix
        ]

        return Vector2D(*values) if len(vector) == 2 else Vector3D(*values)

    def __getitem__(self, index: Tuple[int, int]) -> PyNumber:
        return self._matrix[index[0]][index[1]]

    def __setitem__(self, index: Tuple[int, int], value: PyNumber):
        self._matrix[index[0]][index[1]] = self._dtype(value)

    def as_ndarray(self, dtype: Dtype = FLOAT_DEFAULT) -> ndarray:
        if np is None:
            raise NotImplementedError('numpy is not available')
        return np.array(self._matrix, dtype = dtype)

    @property
    def dtype(self) -> Type:
        return self._dtype

    @classmethod
    def from_ndarray(cls, matrix: np.ndarray, dtype: Type = float) -> MatrixABC:

        assert matrix.ndim == 2
        assert matrix.shape in ((2, 2), (3, 3))

        matrix = matrix.tolist()
        if isinstance(matrix[0][0], int):
            matrix = [[dtype(item) for item in line] for line in matrix]

        return cls(matrix, dtype = dtype)

    @classmethod
    def from_2d_rotation(cls, a: PyNumber) -> MatrixABC:

        return cls([
            [cos(a), -sin(a)],
            [sin(a), cos(a)],
        ])

    @classmethod
    def from_3d_rotation(cls, v: Vector3DABC, a: PyNumber) -> MatrixABC:

        ca = cos(a)
        oca = 1 - ca
        sa = sin(a)

        return cls([
            [ca + (v.x ** 2) * oca, v.x * v.y * oca - v.z * sa, v.x * v.y * oca + v.y * sa],
            [v.y * v.x * oca + v.z * sa, ca + (v.y ** 2) * oca, v.y * v.z * oca - v.x * sa],
            [v.z * v.x * oca - v.y * sa, v.z * v.y * oca + v.x * sa, ca + (v.z ** 2) * oca],
        ])



# -*- coding: utf-8 -*-

"""

BEWEGUNG
a versatile video renderer
https://github.com/pleiszenburg/bewegung

    src/bewegung/core/backends/datashader.py: Datashader backend

    Copyright (C) 2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/bewegung/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from typing import Any, Callable

from PIL.Image import Image
from PIL import ImageOps

from ._base import BackendBase
from ..abc import VideoABC
from ..typeguard import typechecked

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@typechecked
class Backend(BackendBase):

    _name = 'Datashader'

    def __init__(self):

        super().__init__()

        self._DS_Image = None

    def _prototype(self, video: VideoABC, **kwargs) -> Callable:

        if 'plot_width' in kwargs.keys() and 'width' in kwargs.keys():
            kwargs.pop('width')
        if 'plot_height' in kwargs.keys() and 'height' in kwargs.keys():
            kwargs.pop('height')

        if 'plot_width' not in kwargs.keys() and 'width' in kwargs.keys():
            kwargs['plot_width'] = kwargs.pop('width')
        if 'plot_height' not in kwargs.keys() and 'height' in kwargs.keys():
            kwargs['plot_height'] = kwargs.pop('height')

        if 'plot_width' not in kwargs.keys():
            kwargs['plot_width'] = video.width
        if 'plot_height' not in kwargs.keys():
            kwargs['plot_height'] = video.height

        if 'x_range' not in kwargs.keys():
            kwargs['x_range'] = (0, video.width)
        if 'y_range' not in kwargs.keys():
            kwargs['y_range'] = (0, video.height)

        return lambda: self._type(**kwargs)

    def _isinstance(self, obj: Any) -> bool:

        return isinstance(obj, self._DS_Image) # Return type is not a canvas!

    def _load(self):

        from datashader import Canvas as DS_Canvas
        from datashader.transfer_functions import Image as DS_Image

        self._type = DS_Canvas
        self._DS_Image = DS_Image

    def _to_pil(self, obj: Any) -> Image:

        cvs = obj.to_pil()
        if cvs.mode != 'RGBA':
            raise TypeError('unhandled image mode')
        return ImageOps.flip(cvs) # datashader's y axis must be flipped



# -*- coding: utf-8 -*-

"""

BEWEGUNG
a versatile video renderer
https://github.com/pleiszenburg/bewegung

    src/bewegung/core/backends/cairo.py: Cairo backend

    Copyright (C) 2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/bewegung/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from typing import Any, Callable

from PIL.Image import Image, frombuffer, merge

from ._base import BackendBase
from ..abc import VideoABC
from ..typeguard import typechecked

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@typechecked
class Backend(BackendBase):

    _name = 'Cairo'

    def __init__(self):

        super().__init__()

        self._FORMAT_ARGB32 = None
        self._Format = None

    def _prototype(self, video: VideoABC, **kwargs) -> Callable:

        if 'format' not in kwargs.keys():
            kwargs['format'] = self._FORMAT_ARGB32
        if 'width' not in kwargs.keys():
            kwargs['width'] = video.width
        if 'height' not in kwargs.keys():
            kwargs['height'] = video.height

        assert len(kwargs) == 3

        return lambda: self._type(kwargs['format'], kwargs['width'], kwargs['height'])

    def _load(self):

        from cairo import FORMAT_ARGB32, ImageSurface, Format

        self._type = ImageSurface

        self._FORMAT_ARGB32 = FORMAT_ARGB32
        self._Format = Format

    def _to_pil(self, obj: Any) -> Image:

        if obj.get_format() != self._Format.ARGB32:
            raise TypeError('ImageSurface uses unhandled format')

        image = frombuffer(
            mode = 'RGBa',
            size = (obj.get_width(), obj.get_height()),
            data = obj.get_data().tobytes(), # call to "tobytes" required because of RGBa mode
            )
        b, g, r, a = image.split()
        return merge('RGBa', (r, g, b, a)).convert("RGBA")



# -*- coding: utf-8 -*-

"""

BEWEGUNG
a versatile video renderer
https://github.com/pleiszenburg/bewegung

    src/bewegung/core/task.py: Render task

    Copyright (C) 2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/bewegung/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from typing import Any, Callable

from .abc import SequenceABC, TaskABC, TimeABC
from .typeguard import typechecked

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@typechecked
class Task(TaskABC):
    """
    Task for video renderer. Tasks can be ordered.
    """

    def __init__(self,
        sequence: SequenceABC,
        index: int,
        task: Callable,
    ):

        self._sequence = sequence
        self._index = index
        self._task = task

    def __repr__(self) -> str:

        return f'<Task index={self._index:d}>'

    def __call__(self, time: TimeABC) -> Any:

        return self._task(time)

    def __lt__(self, other: TaskABC) -> bool:

        return self.index < other.index

    @property
    def index(self) -> int:

        return self._index

    @property
    def sequence(self) -> SequenceABC:

        return self._sequence



# -*- coding: utf-8 -*-

"""

BEWEGUNG
a versatile video renderer
https://github.com/pleiszenburg/bewegung

    src/bewegung/core/color.py: Simple RGBA Color handling

    Copyright (C) 2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/bewegung/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from typing import Tuple

from .abc import ColorABC
from .typeguard import typechecked

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@typechecked
class Color(ColorABC):
    """
    Common infrastructure for working with RGBA colors in different formats.

    Color objects are immutable.

    Args:
        r : red channel 0...255 (uint8)
        g : green channel 0...255 (uint8)
        b : blue channel 0...255 (uint8)
        a : alpha channel 0...255 (uint8), opaque by default
    """

    def __init__(self,
        r: int,
        g: int,
        b: int,
        a: int = 255,
    ):

        if not (0 <= r <= 255):
            raise ValueError('red color channel out of bounds (0...255)')
        if not (0 <= g <= 255):
            raise ValueError('green color channel out of bounds (0...255)')
        if not (0 <= b <= 255):
            raise ValueError('blue color channel out of bounds (0...255)')
        if not (0 <= a <= 255):
            raise ValueError('alpha color channel out of bounds (0...255)')

        self._r, self._g, self._b, self._a = r, g, b, a

    def __repr__(self) -> str:

        return f'<Color r={self._r:d} g={self._g:d} b={self._b:d} a={self._a:d}>'

    @property
    def r(self) -> int:
        """
        red channel
        """

        return self._r

    @property
    def g(self) -> int:
        """
        green channel
        """

        return self._g

    @property
    def b(self) -> int:
        """
        blue channel
        """

        return self._b

    @property
    def a(self) -> int:
        """
        alpha channel
        """

        return self._a

    def as_hex(self, alpha: bool = True) -> str:
        """
        Exports color as hexadecimal string

        Args:
            alpha : Allows to disable alpha channel on export
        """

        if not alpha:
            return f'{self._r:02x}{self._g:02x}{self._b:02x}'

        return f'{self._r:02x}{self._g:02x}{self._b:02x}{self._a:02x}'

    def as_rgba_float(self) -> Tuple[float, float, float, float]:
        """
        Exports color a tuple of floats 0.0...1.0
        """

        return self._r / 255, self._g / 255, self._b / 255, self._a / 255

    def as_rgba_int(self) -> Tuple[int, int, int, int]:
        """
        Exports color a tuple of ints 0...255 (uint8)
        """

        return self._r, self._g, self._b, self._a

    def as_opaque(self) -> ColorABC:
        """
        Exports color as a new, fully opaque version of itself
        """

        return type(self)(self._r, self._g, self._b, 255)

    def as_transparent(self) -> ColorABC:
        """
        Exports color as a new, fully transparent version of itself
        """

        return type(self)(self._r, self._g, self._b, 0)

    @classmethod
    def from_rgba_float(cls,
        r: float,
        g: float,
        b: float,
        a: float = 1.0,
    ) -> ColorABC:
        """
        Imports color from floats

        Args:
            r : red channel 0.0...1.0
            g : green channel 0.0...1.0
            b : blue channel 0.0...1.0
            a : alpha channel 0.0...1.0, opaque by default
        """

        if not (0.0 <= r <= 1.0):
            raise ValueError('red color channel out of bounds (0.0...1.0)')
        if not (0.0 <= g <= 1.0):
            raise ValueError('green color channel out of bounds (0.0...1.0)')
        if not (0.0 <= b <= 1.0):
            raise ValueError('blue color channel out of bounds (0.0...1.0)')
        if not (0.0 <= a <= 1.0):
            raise ValueError('alpha color channel out of bounds (0.0...1.0)')

        return cls(
            r = round(r * 255),
            g = round(g * 255),
            b = round(b * 255),
            a = round(a * 255),
        )

    @classmethod
    def from_hex(cls, raw: str) -> ColorABC:
        """
        Imports color from a hexadecimal string

        Args:
            raw : Accepts strings both with and without alpha channel. Opaque by default.
        """

        if raw.startswith('#'):
            raw = raw[1:]

        assert len(raw) in (6, 8)

        return cls(**{
            color: int(component, base = 16)
            for color, component in zip(
                ('r', 'g', 'b', 'a'),
                (raw[idx:idx+2] for idx in range(0, 8, 2)),
            )
            if not (color == 'a' and len(component) == 0)
        })



# -*- coding: utf-8 -*-

"""

BEWEGUNG
a versatile video renderer
https://github.com/pleiszenburg/bewegung

    src/bewegung/core/sequence.py: Video Sequence

    Copyright (C) 2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/bewegung/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from typing import Dict

from .abc import SequenceABC, TimeABC, VideoABC
from .typeguard import typechecked

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@typechecked
class Sequence(SequenceABC):
    """
    Holds layers and prepare tasks. "Base class" for user-defined sequences.
    Actually, new classes are created by making this class inherit from user-defined sequence classes.
    Do not instantiate this class or derive from it - use the :meth:`bewegung.Video.sequence` decorator instead!

    Mutable.

    Args:
        start : Begin of sequence within video
        stop : End of sequence within video
        video : Parent video object
    """

    def __init__(self, start: TimeABC, stop: TimeABC, video: VideoABC):

        # consistency checks are performed in Video.sequence

        self._start, self._stop = start, stop
        self._video, self._ctx = video, video.ctx

        self._length = stop - start

    def __repr__(self) -> str:

        return f'<Sequence name={type(self).__name__:s}>'

    def __len__(self) -> int:
        """
        Duration of sequence as number of frames
        """

        return self._length.index

    def __contains__(self, time: TimeABC) -> bool:
        """
        Checks whether a ``Time`` is within the sequence.

        Args:
            time : Time within parent video
        """

        return self._start <= time and time < self._stop

    def reset(self):
        """
        Calls the constructor of the user-defined sequence class.
        Used to reset the sequence for a (new) rendering run.
        """

        super().__init__()

    @property
    def length(self) -> TimeABC:
        """
        Duration of sequence
        """

        return self._length

    @property
    def start(self) -> TimeABC:
        """
        Begin of sequence within video
        """

        return self._start

    @property
    def stop(self) -> TimeABC:
        """
        End of sequence within video
        """

        return self._stop

    @property
    def video(self) -> VideoABC:
        """
        "Parent" video object
        """

        return self._video

    @property
    def ctx(self) -> Dict:
        """
        "Parent" context dictionary
        """

        return self._ctx



# -*- coding: utf-8 -*-

"""

BEWEGUNG
a versatile video renderer
https://github.com/pleiszenburg/bewegung

    src/bewegung/core/vector/array3d.py: 3D Vector Array

    Copyright (C) 2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/bewegung/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from typing import List, Tuple, Union

import numpy as np
from typeguard import typechecked

from .lib import dtype_np2py
from .single3d import Vector3D
from ..abc import Dtype, Number, VectorArray3DABC, VectorIterable3D
from ..const import FLOAT_DEFAULT

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@typechecked
class VectorArray3D(VectorArray3DABC):

    def __init__(self, x: np.ndarray, y: np.ndarray, z: np.ndarray):
        assert x.ndim == 1
        assert y.ndim == 1
        assert z.ndim == 1
        assert x.shape[0] == y.shape[0] == z.shape[0]
        assert x.dtype == y.dtype == z.dtype
        self._x, self._y, self._z = x, y, z

    def __repr__(self) -> str:
        return f'<VectorArray3D len={len(self):d}>'

    def __len__(self) -> int:
        return self._x.shape[0]

    def __getitem__(self, idx: Union[int, slice]) -> Union[Vector3D, VectorArray3DABC]:
        if isinstance(idx, int):
            dtype = dtype_np2py(self.dtype)
            return Vector3D(dtype(self._x[idx]), dtype(self._y[idx]), dtype(self._z[idx]), dtype = dtype)
        return VectorArray3D(self._x[idx].copy(), self._y[idx].copy(), self._z[idx].copy())

    def __eq__(self, other: VectorArray3DABC) -> bool:
        return np.array_equal(self.x, other.x) and np.array_equal(self.y, other.y) and np.array_equal(self.z, other.z)

    def __mod__(self, other: VectorArray3DABC) -> bool:
        return np.allclose(self.x, other.x) and np.allclose(self.y, other.y) and np.allclose(self.z, other.z)

    def __add__(self, other: VectorArray3DABC) -> VectorArray3DABC:
        assert len(self) == len(other)
        assert self.dtype == other.dtype
        return VectorArray3D(self.x + other.x, self.y + other.y, self.z + other.z)

    def __sub__(self, other: VectorArray3DABC) -> VectorArray3DABC:
        assert len(self) == len(other)
        assert self.dtype == other.dtype
        return VectorArray3D(self.x - other.x, self.y - other.y, self.z - other.z)

    def __mul__(self, other: Number) -> VectorArray3DABC:
        return VectorArray3D(self._x * other, self._y * other, self._z * other)

    def mul(self, scalar: Number):
        np.multiply(self._x, scalar, out = self._x)
        np.multiply(self._y, scalar, out = self._y)
        np.multiply(self._z, scalar, out = self._z)

    def __matmul__(self, other: VectorArray3DABC) -> np.ndarray:
        assert len(self) == len(other)
        assert self.dtype == other.dtype
        return self.x * other.x + self.y * other.y + self.z * other.z

    def as_list(self) -> List[Vector3D]:
        dtype = dtype_np2py(self.dtype)
        return [
            Vector3D(dtype(self._x[idx]), dtype(self._y[idx]), dtype(self._z[idx]), dtype = dtype)
            for idx in range(len(self))
        ]

    def as_ndarray(self, dtype: Dtype = FLOAT_DEFAULT) -> np.ndarray:
        a = np.zeros((len(self), 3), dtype = self.dtype)
        a[:, 0], a[:, 1], a[:, 2] = self._x, self._y, self._z
        return a if a.dtype == np.dtype(dtype) else a.astype(dtype)

    def as_polar_tuple(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        return (
            self.mag,
            np.arccos(self._z / self.mag),
            np.arctan2(self._y, self._x),
            )

    def as_tuple(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        return self._x.copy(), self._y.copy(), self._z.copy()

    def as_type(self, dtype: Dtype) -> VectorArray3DABC:
        return self.copy() if self.dtype == np.dtype(dtype) else VectorArray3D(
            self._x.astype(dtype), self._y.astype(dtype), self._z.astype(dtype),
        )

    def copy(self) -> VectorArray3DABC:
        return VectorArray3D(self._x.copy(), self._y.copy(), self._z.copy())

    def update_from_vector(self, other: VectorArray3DABC):
        self._x[:], self._y[:], self._z[:] = other.x[:], other.y[:], other.z[:]

    @property
    def dtype(self) -> np.dtype:
        return self._x.dtype

    @property
    def mag(self) -> np.ndarray:
        return np.sqrt(self._x ** 2 + self._y ** 2 + self._z ** 2)

    @property
    def x(self) -> np.ndarray:
        return self._x
    @x.setter
    def x(self, value: float):
        raise NotImplementedError()

    @property
    def y(self) -> np.ndarray:
        return self._y
    @y.setter
    def y(self, value: float):
        raise NotImplementedError()

    @property
    def z(self) -> np.ndarray:
        return self._z
    @z.setter
    def z(self, value: float):
        raise NotImplementedError()

    @classmethod
    def from_iterable(cls, obj: VectorIterable3D, dtype: Dtype = FLOAT_DEFAULT) -> VectorArray3DABC:
        if not isinstance(obj, list):
            obj = list(obj)
        x = np.zeros((len(obj),), dtype = dtype)
        y = np.zeros((len(obj),), dtype = dtype)
        z = np.zeros((len(obj),), dtype = dtype)
        for idx, item in enumerate(obj):
            x[idx], y[idx], z[idx] = item.x, item.y, item.z
        return cls(x = x, y = y, z = z,)

    @classmethod
    def from_polar(cls, radius: np.ndarray, theta: np.ndarray, phi: np.ndarray) -> VectorArray3DABC:
        assert radius.ndim == 1
        assert theta.ndim == 1
        assert phi.ndim == 1
        assert radius.shape[0] == theta.shape[0] == phi.shape[0]
        assert radius.dtype == theta.dtype == phi.dtype
        RadiusSinTheta = radius * np.sin(theta)
        return cls(
            x = RadiusSinTheta * np.cos(phi),
            y = RadiusSinTheta * np.sin(phi),
            z = radius * np.cos(theta),
            )

    @classmethod
    def from_geographic(cls, radius: np.ndarray, lon: np.ndarray, lat: np.ndarray) -> VectorArray3DABC:
        assert radius.ndim == 1
        assert lon.ndim == 1
        assert lat.ndim == 1
        assert radius.shape[0] == lon.shape[0] == lat.shape[0]
        assert radius.dtype == lon.dtype == lat.dtype
        rad2deg = np.dtype(radius.dtype).type(np.pi / 180.0)
        halfpi = np.dtype(radius.dtype).type(np.pi / 2.0)
        return cls.from_polar(
            radius = radius,
            theta = halfpi - (lat * rad2deg),
            phi = lon * rad2deg,
            )



# -*- coding: utf-8 -*-

"""

BEWEGUNG
a versatile video renderer
https://github.com/pleiszenburg/bewegung

    src/bewegung/__init__.py: Package root

    Copyright (C) 2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/bewegung/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# EXPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

__version__ = '0.0.3'

from .core.camera import Camera
from .core.backends import *
from .core.color import Color
from .core.const import *
from .core.effects import EffectBase, FadeInEffect, FadeOutEffect
from .core.encoders import EncoderBase, FFmpegH264Encoder, FFmpegGifEncoder
from .core.indexpool import IndexPool
from .core.layer import Layer
from .core.sequence import Sequence
from .core.task import Task
from .core.time import Time
from .core.timescale import TimeScale
from .core.vector import *
from .core.video import Video



# -*- coding: utf-8 -*-

"""

BEWEGUNG
a versatile video renderer
https://github.com/pleiszenburg/bewegung

    src/bewegung/core/encoders.py: Wrapper for video encoders

    Copyright (C) 2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/bewegung/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from types import TracebackType
from typing import BinaryIO, Union, Type
from subprocess import Popen, PIPE, DEVNULL

from .abc import EncoderABC, VideoABC
from .const import (
    PIPE_BUFFER_DEFAULT,
    FFMPEG_CRF_DEFAULT,
    FFMPEG_PRESET_DEFAULT,
    FFPMEG_TUNE_DEFAULT,
    )
from .typeguard import typechecked

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS: BASE
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@typechecked
class EncoderBase(EncoderABC):
    """
    Encoder classes wrap video encoding tools and libraries such as ``ffmpeg``.
    Encoder objects are callable and return themselves when called. This mechanism is used to (re-) configure the encoder object.
    Encoder objects also use Python's context manager protocol and expose ``BinaryIO`` objects, i.e. streams, as a context for actual encoding.
    :meth:`bewegung.Video.render` will write rendered images as RGB bitmaps to this stream so the encoder can pick them up.
    Encoder objects can either be "running" or "idling". They can also either be "configured" or "unconfigured".
    In the latter case, they will not allow to encode a video.

    Mutable.

    If the orginal cunstructor method is overridden, it must be called from the child class.
    """

    def __init__(self):

        self._width = None
        self._height = None
        self._fps = None
        self._video_fn = None
        self._stream = None

    def __call__(self, video: VideoABC, video_fn: str) -> EncoderABC:
        """
        Configures the encoder. Returns encoder object itself.

        Args:
            video : Video object
            video_fn : Location and name (path) of where to store the video file.
                If omitted, no video will be rendered.
                However, indivual frames may in fact still be rendered if ``frame_fn`` has been specified.
        """

        if len(video_fn) == 0:
            raise ValueError('video_fn must not be empty')

        self._width = video.width
        self._height = video.height
        self._fps = video.fps
        self._video_fn = video_fn

        return self

    def __repr__(self) -> str:

        return f'<{type(self).__name__} configured={"yes" if self.configured else "no"} running={"yes" if self.running else "no"}>'

    @property
    def configured(self) -> bool:
        """
        Has the encoder been configured?

        Do not override!
        """

        return self._video_fn is not None

    @property
    def running(self) -> bool:
        """
        Is the encoder currently running?

        Do not override!
        """

        return self._stream is not None

    @property
    def stream(self) -> BinaryIO:
        """
        Exposes the input stream of the encoder.
        If the encoder is not running, trying to access this attribute raises an exception.

        Do not override!
        """

        if not self.running:
            raise RuntimeError('encoder is not running')

        return self._stream

    @property
    def video_fn(self) -> str:
        """
        Exposes the name/path of the target video file.
        If the encoder is not configured, trying to access this attribute raises an exception.

        Do not override!
        """

        if not self.configured:
            raise RuntimeError('encoder is not configured')

        return self._video_fn

    @video_fn.setter
    def video_fn(self, value: str):
        """
        Exposes the name/path of the target video file.
        If the encoder is not configured, trying to access this attribute raises an exception.

        Do not override!
        """

        if not self.configured:
            raise RuntimeError('encoder is not configured')
        if self.running:
            raise RuntimeError('encoder is currently running')
        if len(value) == 0:
            raise ValueError('video_fn must not be empty')

        self._video_fn = value

    def __enter__(self) -> BinaryIO:
        """
        Context manager entry point. Returns the encoder's input stream.

        Do not override!
        """

        if not self.configured:
            raise RuntimeError('encoder has not been configured')
        if self.running:
            raise RuntimeError('encoder is already running')

        self._stream = self._enter()
        return self._stream

    def _enter(self) -> BinaryIO:
        """
        Starts the encoder. Returns the encoder's input stream.

        Must be reimplemented!
        """

        raise NotImplementedError()

    def __exit__(
        self,
        exc_type: Union[Type, None],
        exc_value: Union[Exception, None],
        traceback: Union[TracebackType, None],
    ):
        """
        Context manager exit point.

        Do not override!
        """

        if not self.running:
            raise RuntimeError('encoder is not running')

        self._stream = None
        self._exit()

    def _exit(self):
        """
        Stops the encoder.

        Must be reimplemented!
        """

        raise NotImplementedError()

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASSES: ENCODERS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@typechecked
class FFmpegH264Encoder(EncoderBase):
    """
    Mutable. Context manager. Wraps FFmpeg with H.264.

    Args:
        buffersize : Maximum size of buffer in bytes between ``bewegung`` and ``ffmpeg``.
            A larger buffer may have a mildly positive impact on performance.
        preset : ``ffmpeg`` encoding and compression preset. See `ffmpeg's H.264 preset documentation`_ for details.
        crf : ``ffmpeg`` Constant Rate Factor (CRF) value. See `ffmpeg's H.264 CRF documentation`_ for details.
        tune : ``ffmpeg`` tune option. See `ffmpeg's H.264 tune documentation`_ for details.

    .. _`ffmpeg's H.264 preset documentation`: https://trac.ffmpeg.org/wiki/Encode/H.264#Preset
    .. _`ffmpeg's H.264 CRF documentation`: https://trac.ffmpeg.org/wiki/Encode/H.264#crf
    .. _`ffmpeg's H.264 tune documentation`: https://trac.ffmpeg.org/wiki/Encode/H.264#Tune
    """

    def __init__(self,
        buffersize: int = PIPE_BUFFER_DEFAULT,
        preset: str = FFMPEG_PRESET_DEFAULT,
        crf: int = FFMPEG_CRF_DEFAULT,
        tune: str = FFPMEG_TUNE_DEFAULT,
    ):

        super().__init__()

        if buffersize <= 0:
            raise ValueError('buffersize must be greater than 0')
        if preset not in (
            "ultrafast",
            "superfast",
            "veryfast",
            "faster",
            "fast",
            "medium",
            "slow",
            "slower",
            "veryslow",
            ):
            raise ValueError('unknown ffmpeg preset')
        if not (0 <= crf <= 51):
            raise ValueError('ffmpeg crf out of bounds')
        if tune not in (
            "film",
            "animation",
            "grain",
            "stillimage",
            "fastdecode",
            "zerolatency",
            ):
            raise ValueError('unknown ffmpeg tune')

        self._buffersize = buffersize
        self._preset = preset
        self._crf = crf
        self._tune = tune

        self._proc = None

    def _enter(self) -> BinaryIO:

        self._proc = Popen(
            [
                'ffmpeg',
                '-y', # force overwrite of output file
                '-framerate', f'{self._fps:d}',
                '-f', 'image2pipe', # force input format
                '-i', '-', # data from stdin
                '-vcodec', 'bmp', # input codec
                '-s:v', f'{self._width:d}x{self._height:d}',
                '-c:v', 'libx264',
                '-preset', self._preset,
                '-crf', f'{self._crf:d}',
                '-tune', self._tune,
                self._video_fn,
            ],
            stdin = PIPE, stdout = DEVNULL, stderr = DEVNULL,
            bufsize = self._buffersize,
        )

        return self._proc.stdin

    def _exit(self):

        self._proc.stdin.flush()
        self._proc.stdin.close()
        self._proc.wait()

        self._proc = None

@typechecked
class FFmpegGifEncoder(EncoderBase):
    """
    Mutable. Context manager. Wraps FFmpeg with gif.

    Args:
        buffersize : Maximum size of buffer in bytes between ``bewegung`` and ``ffmpeg``.
            A larger buffer may have a mildly positive impact on performance.
    """

    def __init__(self,
        buffersize: int = PIPE_BUFFER_DEFAULT,
    ):

        super().__init__()

        if buffersize <= 0:
            raise ValueError('buffersize must be greater than 0')

        self._buffersize = buffersize

        self._proc = None

    def _enter(self) -> BinaryIO:

        self._proc = Popen(
            [
                'ffmpeg',
                '-y', # force overwrite of output file
                '-framerate', f'{self._fps:d}',
                '-f', 'image2pipe', # force input format
                '-i', '-', # data from stdin
                '-vcodec', 'bmp', # input codec
                '-s:v', f'{self._width:d}x{self._height:d}',
                '-vf', 'split[s0][s1];[s0]palettegen[p];[s1][p]paletteuse',
                '-c:v', 'gif',
                self._video_fn,
            ],
            stdin = PIPE, stdout = DEVNULL, stderr = DEVNULL,
            bufsize = self._buffersize,
        )

        return self._proc.stdin

    def _exit(self):

        self._proc.stdin.flush()
        self._proc.stdin.close()
        self._proc.wait()

        self._proc = None



# -*- coding: utf-8 -*-

"""

BEWEGUNG
a versatile video renderer
https://github.com/pleiszenburg/bewegung

    src/bewegung/core/vector/single2ddist.py: Single 2D Vector with distance parameter

    Copyright (C) 2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/bewegung/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from typeguard import typechecked

from typing import Type, Union

from .single2d import Vector2D
from ..abc import PyNumber, Vector2DABC

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@typechecked
class Vector2Ddist(Vector2D):
    """
    Immutable version of Vector2D with distance parameter
    """

    def __init__(self, x: PyNumber, y: PyNumber, dist: PyNumber, dtype: Union[Type, None] = None):
        super().__init__(x = x, y = y, dtype = dtype)
        assert isinstance(dist, self._dtype)
        self._dist = dist

    def __repr__(self) -> str:
        if self._dtype == int:
            return f'<Vector2Ddist x={self._x:d} y={self._y:d} dist={self._dist:d} dtype={self._dtype.__name__:s}>'
        return f'<Vector2Ddist x={self._x:e} y={self._y:e} dist={self._dist:e} dtype={self._dtype.__name__:s}>'

    def mul(self, scalar: PyNumber):
        raise NotImplementedError()

    def as_vector(self) -> Vector2DABC:
        return Vector2D(self._x, self._y, dtype = self._dtype)

    def copy(self) -> Vector2DABC:
        return type(self)(self._x, self._y, self._dist, dtype = self._dtype)

    def update(self, x: PyNumber, y: PyNumber):
        raise NotImplementedError()

    def update_from_vector(self, other: Vector2DABC):
        raise NotImplementedError()

    @property
    def x(self) -> PyNumber:
        return self._x
    @x.setter
    def x(self, value: PyNumber):
        raise NotImplementedError()

    @property
    def y(self) -> PyNumber:
        return self._y
    @y.setter
    def y(self, value: PyNumber):
        raise NotImplementedError()

    @property
    def dist(self) -> PyNumber:
        return self._dist
    @dist.setter
    def dist(self, value: PyNumber):
        raise NotImplementedError()

    @classmethod
    def from_polar(cls, radius: PyNumber, angle: PyNumber) -> Vector2DABC:
        raise NotImplementedError()



# -*- coding: utf-8 -*-

"""

BEWEGUNG
a versatile video renderer
https://github.com/pleiszenburg/bewegung

    src/bewegung/contrib/circular_century_calendar.py: Circular Century Calendar

    Copyright (C) 2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/bewegung/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import calendar
from datetime import datetime
import math
from typing import List, Union

from PIL.Image import Image, new, LANCZOS

from ..core.canvas import inventory
from ..core.color import Color
from ..core.vector import Matrix, Vector2D
from ..core.typeguard import typechecked

DrawingBoard = inventory['drawingboard'].type

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@typechecked
class _Background(DrawingBoard):

    _twopi = 2 * math.pi
    _halfpi = math.pi / 2

    def draw_ticks(self,
        r1: float,
        r2: float,
        ticks: range,
        line_color: Color,
        line_width: float,
        zero: float = 0.0,
        length: Union[float, None] = None
    ):

        if length is None:
            length = len(ticks)

        af = self._twopi / length

        for tick in ticks:
            angle = (tick - zero) * af - self._halfpi
            self.draw_polygon(
                Vector2D.from_polar(r1, angle),
                Vector2D.from_polar(r2, angle),
                line_color = line_color,
                line_width = line_width,
            )

    def draw_labels(self,
        r: float,
        labels: List[str],
        font: DrawingBoard.Pango.FontDescription,
        font_color: Color,
        zero: float = 0.0,
        length: Union[float, None] = None
    ):

        if length is None:
            length = len(labels)

        af = self._twopi / length

        for idx, label in enumerate(labels):
            angle = (idx - zero) * af - self._halfpi
            self.draw_text(
                text = label,
                point = Vector2D.from_polar(r, angle),
                font = font,
                font_color = font_color,
            )

@typechecked
class _Foreground(DrawingBoard):

    _twopi = 2 * math.pi
    _days_in_decade = {}

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        self._days_in_decade.update({
            decade: sum([
                self._days_in_year(decade + year)
                for year in range(10)
            ])
            for decade in range(1970, 2070 + 10, 10)
        })

    def draw_hands(self, dt: datetime, factor: float, color: Color):

        mr = Matrix.from_2d_rotation(self._angle_from_day_in_century(dt))
        self.draw_filledpolygon(
            mr @ (Vector2D(-7.0, -98.0) * factor),
            mr @ (Vector2D(0, -118.0) * factor),
            mr @ (Vector2D(7.0, -98.0) * factor),
            fill_color = color,
        ) # decade in century

        mr = Matrix.from_2d_rotation(self._angle_from_day_in_decade(dt))
        self.draw_filledpolygon(
            mr @ (Vector2D(-7.0, -160.0) * factor),
            mr @ (Vector2D(0, -180.0) * factor),
            mr @ (Vector2D(7.0, -160.0) * factor),
            fill_color = color,
        ) # year in decade

        mr = Matrix.from_2d_rotation(self._angle_from_day_in_year(dt))
        self.draw_filledcircle(
            point = mr @ (Vector2D(0.0, -218.0) * factor),
            r = 8,
            fill_color = color,
        )

    @staticmethod
    def _days_in_year(year: int) -> int:
        return 366 if calendar.isleap(year) else 365

    @staticmethod
    def _decade_from_year(year: int):
        return int(year / 10) * 10

    @staticmethod
    def _day_in_year(dt: datetime):
        return dt.timetuple().tm_yday

    @classmethod
    def _day_in_decade(cls, dt: datetime):
        return sum([
            cls._days_in_year(previous_year)
            for previous_year in range(cls._decade_from_year(dt.year), dt.year)
        ]) + cls._day_in_year(dt)

    @classmethod
    def _angle_from_day_in_year(cls, dt: datetime) -> float:
        return cls._twopi * cls._day_in_year(dt) / cls._days_in_year(dt.year)

    @classmethod
    def _angle_from_day_in_decade(cls, dt: datetime) -> float:
        return cls._twopi * cls._day_in_decade(dt) / cls._days_in_decade[cls._decade_from_year(dt.year)]

    @classmethod
    def _angle_from_day_in_century(cls, dt: datetime) -> float:
        fraction = cls._day_in_decade(dt) / cls._days_in_decade[cls._decade_from_year(dt.year)]
        fraction += math.floor((dt.year - 2020) / 10)
        return 10 * fraction * cls._twopi / 170

@typechecked
class CircularCenturyCalendar:

    def __init__(self,
        side: int = 450,
        subpixels: int = 1,
        background_color: Union[Color, None] = None,
        foreground_color: Union[Color, None] = None,
    ):

        if background_color is None:
            background_color = Color(0, 0, 0)
        if foreground_color is None:
            foreground_color = Color(255, 0, 0)

        self._actual_side = side
        self._subpixels = subpixels
        self._background_color = background_color
        self._foreground_color = foreground_color

        self._side = self._actual_side * subpixels
        self._center = Vector2D(self._side / 2, self._side / 2)
        self._factor = self._side / 450

        self._font_years = DrawingBoard.make_font("Oxygen Mono", 20.0 * self._factor)
        self._font_decades = DrawingBoard.make_font("Oxygen Mono", 10.0 * self._factor)
        self._font_date = DrawingBoard.make_font("Oxygen Mono", 35.0 * self._factor)

        self._background = self._draw_background()

    def __call__(self, dt: datetime) -> Image:

        foreground = _Foreground(
            width = self._side,
            height = self._side,
            offset = self._center,
            background_color = self._background_color.as_transparent(),
        )

        foreground.draw_text(
            text = dt.strftime('%Y-%m-%d'),
            point = Vector2D(0.0, 90.0 * self._factor),
            font = self._font_date,
            font_color = self._foreground_color,
        )

        foreground.draw_hands(
            dt = dt,
            factor = self._factor,
            color = self._foreground_color,
        )

        foreground = foreground.as_pil()

        image = new(
            'RGBA',
            (self._side, self._side),
            self._background_color.as_transparent().as_rgba_int(),
        ) # transparent black
        image.paste(im = self._background, mask = self._background)
        image.paste(im = foreground, mask = foreground)

        return image.resize(
            (self._actual_side, self._actual_side),
            resample = LANCZOS,
        )

    def _draw_background(self) -> Image:

        canvas = _Background(
            width = self._side,
            height = self._side,
            offset = self._center,
            background_color = self._background_color,
            )

        canvas.draw_ticks(
            r1 = 210.0 * self._factor,
            r2 = 193.0 * self._factor,
            ticks = range(0, 10 * 12),
            line_color = self._foreground_color,
            line_width = 2.0 * self._factor,
            ) # months in decade (outer minor ticks)
        canvas.draw_ticks(
            r1 = 210.0 * self._factor,
            r2 = 187.0 * self._factor,
            ticks = range(0, 10),
            line_color = self._foreground_color,
            line_width = 5.0 * self._factor,
            ) # years in decade (outer major ticks)
        canvas.draw_labels(
            r = 169 * self._factor,
            labels = [f'{year:d}' for year in range(10)],
            font = self._font_years,
            font_color = self._foreground_color,
        ) # years in decade (outer major ticks labels)

        canvas.draw_ticks(
            r1 = 145.0 * self._factor,
            r2 = 128.0 * self._factor,
            ticks = range(1970, 2070 + 2, 2),
            line_color = self._foreground_color,
            line_width = 1.0 * self._factor,
            zero = 2020,
            length = 170,
            ) # year in century (inner minor ticks)
        canvas.draw_ticks(
            r1 = 145.0 * self._factor,
            r2 = 124.0 * self._factor,
            ticks = range(1970, 2070 + 10, 10),
            line_color = self._foreground_color,
            line_width = 2.5 * self._factor,
            zero = 2020,
            length = 170,
            ) # decade in century (inner major ticks)
        labels = [f'{decade:d}' for decade in range(1970, 2070 + 10, 10)]
        canvas.draw_labels(
            r = 102 * self._factor,
            labels = labels,
            font = self._font_decades,
            font_color = self._foreground_color,
            zero = labels.index('2020'),
            length = 17,
        ) # decade in century (inner major ticks labels)

        return canvas.as_pil()



# -*- coding: utf-8 -*-

"""

BEWEGUNG
a versatile video renderer
https://github.com/pleiszenburg/bewegung

    setup.py: Used for package distribution

    Copyright (C) 2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/bewegung/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""


# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from setuptools import (
    find_packages,
    setup,
)
import os

from docs.version import get_version

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# SETUP
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

# List all versions of Python which are supported
python_minor_min = 6
python_minor_max = 8
confirmed_python_versions = [
    "Programming Language :: Python :: 3.{MINOR:d}".format(MINOR=minor)
    for minor in range(python_minor_min, python_minor_max + 1)
]

# Fetch readme file
with open(os.path.join(os.path.dirname(__file__), "README.md")) as f:
    long_description = f.read()

# Define source directory (path)
SRC_DIR = "src"

# Version
__version__ = get_version()

# Requirements
extras_require = {
    "dev": [
        "black",
        "python-language-server[all]",
        "psutil",
        "setuptools",
        "Sphinx",
        "sphinx-autodoc-typehints",
        "sphinx-rtd-theme",
        "sphinxembeddedvideos", # https://github.com/sphinx-contrib/youtube/issues/9#issuecomment-734295832
        "twine",
        "wheel",
    ],
    "drawingboard": [
        "pycairo",
        "PyGObject",
    ],
    "cairo": [
        "pycairo",
    ],
    "datashader": [
        "datashader", # datashader backend (optional)
    ],
    "matplotlib": [
        "numpy",
        "matplotlib", # matplotlib backend (optional)
        "pycairo",
        "mplcairo",
    ],
    "ipython": [
        "ipython", # for drawingboard backend (optional)
    ],
    "numba": [
        "numba", # for camera (optional)
    ],
    "numpy": [
        "numpy", # for camera (optional) and vector arrays (required)
    ],
    "tqdm": [
        "tqdm", # for progress bar during video rendering (optional)
    ],
    "typeguard": [
        "typeguard", # for type checking (optional)
    ],
}
extras_require["all"] = list(
    {rq for target in extras_require.keys() for rq in extras_require[target]}
)
extras_require["docs"] = list(set(extras_require["all"]) - {'PyGObject'})

# Install package
setup(
    name="bewegung",
    packages=find_packages(SRC_DIR),
    package_dir={"": SRC_DIR},
    version=__version__,
    description="a versatile video renderer",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Sebastian M. Ernst",
    author_email="ernst@pleiszenburg.de",
    url="https://github.com/pleiszenburg/bewegung",
    download_url="https://github.com/pleiszenburg/bewegung/archive/v%s.tar.gz"
    % __version__,
    license="LGPLv2",
    keywords=[
        "video",
        "animation",
        "animation-framework",
        "rendering",
        "rendering",
        "rendering-engine",
        "compositor",
    ],
    scripts=[],
    include_package_data=True,
    python_requires=">=3.{MINOR:d}".format(MINOR=python_minor_min),
    setup_requires=[],
    install_requires=[
        "Pillow",
        ],
    extras_require=extras_require,
    zip_safe=False,
    entry_points={},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Financial and Insurance Industry",
        "Intended Audience :: Healthcare Industry",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: BSD",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
    ]
    + confirmed_python_versions
    + [
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Artistic Software",
        "Topic :: Education",
        "Topic :: Multimedia",
        "Topic :: Multimedia :: Graphics",
        "Topic :: Multimedia :: Graphics :: Presentation",
        "Topic :: Multimedia :: Sound/Audio",
        "Topic :: Multimedia :: Sound/Audio :: Editors",
        "Topic :: Multimedia :: Sound/Audio :: Mixers",
        "Topic :: Multimedia :: Video",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Software Development :: Libraries",
    ],
)



# -*- coding: utf-8 -*-

"""

BEWEGUNG
a versatile video renderer
https://github.com/pleiszenburg/bewegung

    src/bewegung/core/vector/single3d.py: Single 3D Vector

    Copyright (C) 2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/bewegung/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

import math
from typing import Tuple, Type, Union

try:
    import numpy as np
    from numpy import ndarray
except ModuleNotFoundError:
    np, ndarray = None, None
from typeguard import typechecked

from ..abc import Dtype, PyNumber, PyNumber3D, Vector3DABC
from ..const import FLOAT_DEFAULT

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@typechecked
class Vector3D(Vector3DABC):

    _rad2deg = math.pi / 180.0
    _halfpi = math.pi / 2.0

    def __init__(self, x: PyNumber, y: PyNumber, z: PyNumber, dtype: Union[Type, None] = None):

        assert type(x) == type(y) == type(z)
        if dtype is None:
            dtype = type(x)
        else:
            assert dtype == type(x)

        self._x, self._y, self._z, self._dtype = x, y, z, dtype

    def __repr__(self) -> str:
        if self._dtype == int:
            return f'<Vector3D x={self._x:d} y={self._y:d} z={self._z:d} dtype={self._dtype.__name__:s}>'
        return f'<Vector3D x={self._x:e} y={self._y:e} z={self._z:e} dtype={self._dtype.__name__:s}>'

    def __eq__(self, other: Vector3DABC) -> bool:
        return (self.x == other.x) and (self.y == other.y) and (self.z == other.z)

    def __mod__(self, other: Vector3DABC) -> bool:
        return math.isclose(self.x, other.x) and math.isclose(self.y, other.y) and math.isclose(self.z, other.z)

    def __add__(self, other: Vector3DABC) -> Vector3DABC:
        return type(self)(self.x + other.x, self.y + other.y, self.z + other.z)

    def __sub__(self, other: Vector3DABC) -> Vector3DABC:
        return type(self)(self.x - other.x, self.y - other.y, self.z - other.z)

    def __mul__(self, other: PyNumber) -> Vector3DABC:
        return type(self)(self._x * other, self._y * other, self._z * other)

    def mul(self, scalar: PyNumber):
        self._x *= scalar
        self._y *= scalar
        self._z *= scalar
        assert type(self._x) == type(self._y) == type(self._z)
        self._dtype = type(self._x)

    def __matmul__(self, other: Vector3DABC) -> PyNumber:
        return self.x * other.x + self.y * other.y + self.z * other.z

    def as_dtype(self, dtype: Type) -> Vector3DABC:
        if dtype == self._dtype:
            return self.copy()
        return type(self)(dtype(self._x), dtype(self._y), dtype(self._z), dtype)

    def as_ndarray(self, dtype: Dtype = FLOAT_DEFAULT) -> ndarray:
        if np is None:
            raise NotImplementedError('numpy is not available')
        return np.array(self.as_tuple(), dtype = dtype)

    def as_polar_tuple(self) -> Tuple[float, float, float]:
        return (
            self.mag,
            math.acos(self._z / self.mag),
            math.atan2(self._y, self._x),
            )

    def as_tuple(self) -> PyNumber3D:
        return (self._x, self._y, self._z)

    def copy(self) -> Vector3DABC:
        return type(self)(self._x, self._y, self._z, self._dtype)

    def update(self, x: PyNumber, y: PyNumber, z: PyNumber):
        assert type(x) == type(y) == type(z)
        self._x, self._y, self._z = x, y, z
        self._dtype = type(self._x)

    def update_from_vector(self, other: Vector3DABC):
        assert type(other.x) == type(other.y) == type(other.z)
        self._x, self._y, self._z = other.x, other.y, other.z
        self._dtype = type(self._x)

    @property
    def mag(self) -> float:
        return math.sqrt(self._x ** 2 + self._y ** 2 + self._z ** 2)

    @property
    def x(self) -> PyNumber:
        return self._x
    @x.setter
    def x(self, value: PyNumber):
        assert isinstance(value, self._dtype)
        self._x = value

    @property
    def y(self) -> PyNumber:
        return self._y
    @y.setter
    def y(self, value: PyNumber):
        assert isinstance(value, self._dtype)
        self._y = value

    @property
    def z(self) -> PyNumber:
        return self._z
    @z.setter
    def z(self, value: PyNumber):
        assert isinstance(value, self._dtype)
        self._z = value

    @property
    def dtype(self) -> Type:
        return self._dtype

    @classmethod
    def from_polar(cls, radius: PyNumber, theta: PyNumber, phi: PyNumber) -> Vector3DABC:
        RadiusSinTheta = radius * math.sin(theta)
        return cls(
            x = RadiusSinTheta * math.cos(phi),
            y = RadiusSinTheta * math.sin(phi),
            z = radius * math.cos(theta),
            )

    @classmethod
    def from_geographic(cls, radius: PyNumber, lon: PyNumber, lat: PyNumber) -> Vector3DABC:
        return cls.from_polar(
            radius = radius,
            theta = cls._halfpi - (lat * cls._rad2deg),
            phi = lon * cls._rad2deg,
            )



# -*- coding: utf-8 -*-

"""

BEWEGUNG
a versatile video renderer
https://github.com/pleiszenburg/bewegung

    src/bewegung/core/layer.py: Layer function/method wrapper

    Copyright (C) 2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/bewegung/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from types import MethodType
from typing import Any, Callable, Union

from PIL import Image as PIL_Image

from .abc import EffectABC, LayerABC, SequenceABC, TimeABC, VideoABC, Vector2DABC
from .backends import backends
from .typeguard import typechecked
from .vector import Vector2D

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@typechecked
class Layer(LayerABC):
    """
    Callable layer method wrapper (decorator), managing rendering and application of effects.
    Do not instantiate this class or derive from it - use the :meth:`bewegung.Video.layer` decorator instead!

    Mutable.

    Args:
        method : Wrapped layer method from a user-defined sequence class
        zindex : A number, managed by an index pool, representing the relative position within a stack of ``layer`` tasks.
        video : Parent video object
        canvas : A function pointer, generating a new canvas once per frame for the ``layer`` task.
        offset : The layer's offset relative to the top-left corner of the video. The y-axis is downwards positive.
    """

    def __init__(self,
        method: Callable,
        zindex: int,
        video: VideoABC,
        canvas: Union[Callable, None] = None,
        offset: Union[Vector2DABC, None] = None,
    ):

        # consistency checks are performed in Video.layer

        if offset is None:
            offset = Vector2D(0, 0)

        self._method = method
        self._zindex_tag = zindex
        self._video = video
        self._canvas = self._video.canvas() if canvas is None else canvas
        self._offset = offset
        self._effects = []

        self._args = self._method.__code__.co_varnames[
            1:self._method.__code__.co_argcount # excluding self and internal namespace
            ] # parameters requested by user

    def __repr__(self) -> str:

        return f'<Layer name={self._method.__name__:s} zindex={self._zindex_tag:d}>'

    def __call__(self, sequence: SequenceABC, time: TimeABC) -> PIL_Image.Image:
        """
        Wraps layer method from a user-defined sequence class.
        This method determines what parameters the user-defined layer method requested.
        Possible options are:

        - ``time``: The absolute time within the parent video
        - ``reltime``: The relative time within the parent sequence
        - ``canvas``: An empty canvas

        Subsequently, the user-defined layer method is called with the requested parameters.
        It then converts the returned canvas to a Pillow Image object
        by making the currently loaded backends recognize the returned canvas type.
        Finally, effects are applied and the Image object is returned.

        Args:
            sequence : Parent sequence
            time : Time within video
        """

        kwargs = {}
        cvs_start = None

        for param in self._args:
            if param == 'time':
                kwargs[param] = time
            elif param == 'reltime':
                kwargs[param] = time - sequence.start
            elif param == 'canvas':
                kwargs[param] = self._canvas()
                cvs_start = kwargs[param]
            else:
                raise ValueError('unknown parameter')

        cvs = self._method(sequence, **kwargs)
        if cvs is None:
            if cvs_start is not None:
                cvs = cvs_start
            else:
                raise ValueError('layer is missing a canvas')
        cvs = self._to_pil(cvs)

        for effect in self._effects:
            cvs = effect.apply_(
                cvs = cvs,
                video = self._video,
                sequence = sequence,
                time = time,
            )

        cvs.offset = self._offset # annotate offset for later use

        return cvs

    def __get__(self, obj, objtype = None):
        # Simulate ``func_descr_get()`` in ``Objects/funcobject.c``, see `stackoverflow`_.
        # .. _stackoverflow: https://stackoverflow.com/q/26226604/1672565

        if obj is None:
            return self

        return MethodType(self, obj)

    @property
    def zindex_tag(self) -> int:
        """
        z-index of layer
        """

        return self._zindex_tag

    def register_effect(self, effect: EffectABC):
        """
        Interface used by effects decorators to register themselves. See :meth:`bewegung.EffectBase.__call__`.

        Args:
            effect : Configured effect object
        """

        self._effects.append(effect)

    def _to_pil(self, obj: Any) -> PIL_Image.Image:
        """
        Detects the datatype of the canvas returned by the user-defined layer method and tries to convert it to a Pillow Image.
        Raises a type error if none of the currently loaded backends recognizes the canvas type.

        Args:
            obj : A canvas object
        """

        for backend in backends.values():
            if backend.isinstance(obj, hard = False):
                return backend.to_pil(obj)

        raise TypeError('unknown or unloaded backend canvas type coming from layer')



# -*- coding: utf-8 -*-

"""

BEWEGUNG
a versatile video renderer
https://github.com/pleiszenburg/bewegung

    src/bewegung/core/backends/_load.py: Backend inventory and loader

    Copyright (C) 2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/bewegung/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from typing import Any

import importlib
import os

from ..typeguard import typechecked

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@typechecked
class _Inventory(dict):
    """
    Backend inventory (a dict, more or less).
    Mutable.
    """

    def __init__(self):

        super().__init__()
        path = os.path.dirname(__file__)
        backend_modules = [
            item[:-3] if item.lower().endswith('.py') else item
            for item in os.listdir(path)
            if not item.startswith('_') and not item.startswith('.')
            ]

        for name in backend_modules:
            self[name] = importlib.import_module(f'bewegung.core.backends.{name:s}').Backend()

    def isinstance(self, obj: Any) -> bool:

        return any((backend.isinstance(obj) for backend in self.values()))

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# EXPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

backends = _Inventory()



# -*- coding: utf-8 -*-

"""

BEWEGUNG
a versatile video renderer
https://github.com/pleiszenburg/bewegung

    src/bewegung/core/vector/lib.py: Vector library

    Copyright (C) 2020 Sebastian M. Ernst <ernst@pleiszenburg.de>

<LICENSE_BLOCK>
The contents of this file are subject to the GNU Lesser General Public License
Version 2.1 ("LGPL" or "License"). You may not use this file except in
compliance with the License. You may obtain a copy of the License at
https://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
https://github.com/pleiszenburg/bewegung/blob/master/LICENSE

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for the
specific language governing rights and limitations under the License.
</LICENSE_BLOCK>

"""

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# IMPORT
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

from typing import Type

import numpy as np
from typeguard import typechecked

from ..abc import Dtype

# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
# CLASS
# +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

@typechecked
def dtype_np2py(dtype: Dtype) -> Type:

    if np.issubdtype(dtype, np.integer):
        return int
    elif np.issubdtype(dtype, np.floating):
        return float
    else:
        raise TypeError("numpy dtype can not be mapped on Python number types")
